from gevent.queue import Queue, Empty
import boto3
from . import config
from .config import IS_DEV
#import umysql

conn_pools = {}
boto_sessions = {
    0: None # 0 is the default boto session id
}

def get_dynamodb_conn():
    if(IS_DEV):
        return boto3.resource(
            'dynamodb',
            endpoint_url='http://{endpoint}:{port}'.format(endpoint="127.0.0.1", port="8000"),
            aws_access_key_id="",
            aws_secret_access_key="",
            region_name="ap-south-1"
        )
    else:
        return boto_sessions[0].resource('dynamodb')

def get_s3_conn():
    return boto_sessions[0].resource('s3')

def get_sqs_conn():
    return boto_sessions[0].client('sqs')

def get_ses_conn():
    return boto_sessions[0].client('ses')


#default connection generators
_pool_item_generators = {
    "dynamodb": get_dynamodb_conn,
    "s3" : get_s3_conn,
    "sqs": get_sqs_conn,
    "ses": get_ses_conn
}

def register_pool_item_generator(pool_name, func):
    _pool_item_generators[pool_name] = func


def get_from_pool(pool_id):
    # check and initialize
    if(boto_sessions[0] == None and config.aws_config):
        boto_sessions[0] = boto3.session.Session(**config.aws_config)

    conn = None
    conn_pool = conn_pools.get(pool_id, None)
    if (conn_pool == None):
        #if no conn_pool for given pool_id create and store in conn_pools
        conn_pool = Queue()
        conn_pools[pool_id] = conn_pool

    try:
        conn = conn_pool.get(block=False)
    # print "reusing from pool"
    except Empty:
        #if queue is empty create a connection for use.
        if(pool_id in _pool_item_generators):
            conn = _pool_item_generators[pool_id]()

    return conn


def release_to_pool(conn, pool_id):
    # print "releasing to pool"
    conn_pools[pool_id].put(conn)


def use_connection_pool(**pool_args):
    def use_db_connection(func):
        def wrapper(*args, **kwargs):

            conn_args = {}
            for pool_arg, pool_id in pool_args.items():
                conn = get_from_pool(pool_id)
                conn_args[pool_arg] = conn
            ret = None
            conn_args.update(kwargs)
            ret = func(*args, **conn_args)
            for pool_arg, pool_id in pool_args.items():
                release_to_pool(conn_args[pool_arg], pool_id)
            return ret

            # TODO: try to remove connection timeout
            # retry = 0
            # while (retry < 2):
            #     conn_args = {}
            #     for pool_arg, pool_id in pool_args.iteritems():
            #         conn = get_from_pool(pool_id)
            #         conn_args[pool_arg] = conn
            #     ret = None
            #     try:
            #         conn_args.update(kwargs)
            #         ret = func(*args, **conn_args)
            #         for pool_arg, pool_id in pool_args.iteritems():
            #             release_to_pool(conn_args[pool_arg], pool_id)
            #         return ret
            #     except Exception, e:
            #         logger.error(message=e, event="message") # dont release that stupid connections to pool, instread write destroyers for each pool
            #         logger.error(message=traceback.format_exc(), event="message")
            #
            #         retry += 1
            #         if retry >= 2:
            #             raise e
        wrapper._original = getattr(func, "_original", func)
        return wrapper

    return use_db_connection


'''
uses gevent.Queue to maintain pools

@use_connection_pool(arg="pool_name")
def func(arg=None):
    # arg is automatically populated from pool_name,
    # and after function is executed it is put back to pool
    pass

or

def func():
    get_from_pool("pool_name")
    release_to_pool("pool_name")


'''



from gevent import monkey
monkey.patch_all()
from gevent import pywsgi
from wheezy.http import HTTPResponse
from wheezy.http import WSGIApplication
from wheezy.routing import url
from wheezy.web.handlers import BaseHandler
from wheezy.web.middleware import bootstrap_defaults
from wheezy.web.middleware import path_routing_middleware_factory

def test(request):
    response = HTTPResponse()
    response.write('<html><body>Hello World</body></html>')
    return response


routes = [
    url('test', test, name='test')
]


options = {}
main = WSGIApplication(
    middleware=[
        bootstrap_defaults(url_mapping=routes),
        path_routing_middleware_factory
    ],
    options=options
)


server = pywsgi.WSGIServer(('127.0.0.1', 5000), main, backlog=128000)
server.serve_forever()




import timeit

import_module = '''from io import BytesIO'''
code = '''
a = bytearray()
i = 0
while(i < 100000):
	a.extend(b'abcd')
	i+=1
'''
code2 = '''
a = ''
i = 0
while(i < 100000):
	a += 'abcd'
	i+=1
'''
code3 = '''
a = b''
i = 0
while(i < 100000):
	a += b'abcd'
	i+=1
'''
code4 = '''
a = BytesIO()
i = 0
while(i < 100000):
	a.write(b'abcd')
	i+=1
a = a.getvalue()
'''

print(timeit.repeat(stmt=code, repeat=5, number=100))
print(timeit.repeat(stmt=code2, repeat=5, number=100))
#print(timeit.repeat(stmt=code3, repeat=5, number=100))
print(timeit.repeat(stmt=code4, setup=import_module, repeat=5, number=100))



import typing
import re
from base64 import b64decode
from datetime import datetime
import ujson as json
from typing import get_type_hints
from functools import partial
#bare minimum validations and schema generators
_OBJ_END_ = object()



class _Optional:
	def __init__(self, *types):
		self._types = types

	def __getitem__(self, *keys):
		return _Optional(*keys)


Optional = _Optional()

class Int:
	def __init__(self, one_of=None, _min=_OBJ_END_, _max=_OBJ_END_, default=_OBJ_END_):
		self._min = _min
		self._max = _max
		self.one_of = set(one_of) if one_of else None
		self._default = default
		#make schema
		self._schema_ = _schema = {"type": "integer"}
		if(default != _OBJ_END_):
			_schema["default"] = default
		if(self._min != _OBJ_END_):
			_schema["minimum"] = self._min
		if(self._max != _OBJ_END_):
			_schema["maximum"] = self._max
		if(self.one_of):
			_schema["enum"] = list(self.one_of)

	def validate(self, e):
		if(e == None):
			if(self._default != _OBJ_END_):
				return self._default
			raise Exception("should be int")
		if(self._min != _OBJ_END_ and e < self._min):
			raise Exception("should be minlen %d"%(self._min))
		if(self._max != _OBJ_END_ and e > self._max):
			raise Exception("more than maxlen %d"%(self._max))
		if(self.one_of and e not in self.one_of):
			raise Exception("should be one of %s"%(str(self.one_of)))
		return e

class Str:

	format_validators = {
		"date-time": lambda e: datetime.strptime(e.strip(), "%Y-%m-%dT%H:%M:%S.%fZ"),
		"date": lambda e: datetime.strptime(e.strip(), "%Y-%m-%d"),
		"binary": lambda e: e,
		"byte": lambda e: b64decode(e)
	}

	def __init__(self, one_of=None, minlen=0, maxlen=4294967295, regex=None, default=_OBJ_END_, **kwargs):
		self.minlen = minlen
		self.maxlen = maxlen
		self.one_of = set(one_of) if one_of else None
		self._default = default
		self.regex = regex and re.compile(regex)
		_fmt = kwargs.pop("format", None)
		self.fmt = _fmt and Str.format_validators[_fmt]

		#make schema
		self._schema_ = _schema = {"type": "string"}
		if(default != _OBJ_END_):
			_schema["default"] = default
		_schema = {"type": "string"}
		if(self.minlen):
			_schema["minimum"] = self.minlen
		if(self.maxlen and self.maxlen < 4294967295):
			_schema["maximum"] = self.maxlen
		if(self.one_of):
			_schema["enum"] = list(self.one_of)
		if(regex):
			_schema["pattern"] = regex
		if(_fmt):
			_schema["format"] = _fmt

	def validate(self, e):
		if(e == None):
			if(self._default != _OBJ_END_):
				return self._default
			raise Exception("should be string")
		if(len(e) < self.minlen):
			raise Exception("should be minlen %s"%(self.minlen))
		if(len(e) > self.maxlen):
			e = e[:self.maxlen]
		if(self.one_of and e not in self.one_of):
			raise Exception("should be one of %s"%(str(self.one_of)))
		if(self.regex and not self.regex.fullmatch(e)):
			raise Exception("did not match the pattern %s"%(str(self.one_of)))
		if(self.fmt):
			return self.fmt(e)
		return e

class Array:
	def __init__(self, _types, default=_OBJ_END_):
		self._types = _types if isinstance(_types, (list, tuple)) else [_types]
		self._default = default

		#ceontents validation
		_s, _v = schema(self._types)

		self._schema_ = _schema = {"type": "array", "items": _s}
		if(default != _OBJ_END_):
			_schema["default"] = default
		#this is spefic validataion
		self.validate = partial(
			array_validation,
			mix=isinstance(self._types, tuple),
			default=self._default,
			**_v.keywords
		)


class Object:
	def __init__(self, default=_OBJ_END_, _required_=None, **keys):
		self._keys = keys
		self._default = default
		self._required = set(_required_) if _required_ else set()

		#instance specific schema
		self._properties = {}
		self._validations = {}
		for k, _type in keys.items():
			_s, _v = schema(_type)
			self._properties[k] = _s
			self._validations[k] = _v

		self._schema_ = _schema = {"type": "object", "properties": self._properties}
		if(default != _OBJ_END_):
			_schema["default"] = default

	#validates Object(a=str, b=int, h=Test)
	def validate(self, e=_OBJ_END_):
		is_validating_dict = True
		if(e == _OBJ_END_):
			e = self.__dict__
			is_validating_dict = False
		elif(isinstance(e, str)):
			e = json.loads(e)

		k = None
		attr_value = None
		cls = self.__class__
		try:
			for k, attr_validation in self._validations.items():
				attr_value = e.get(k, _OBJ_END_)
				if(attr_value != _OBJ_END_):
					_validated_attr = attr_validation(attr_value)
				else:
					_validated_attr = getattr(cls, k, None) # try getting class attribute dfault

				if(_validated_attr == None and k in self._required):
					raise Exception("Field is required")

				if(_validated_attr != attr_value):
					e[k] = _validated_attr
			return e if is_validating_dict else self
		except Exception as ex:
			raise Exception({
				"exception": (ex.args and ex.args[0]) or "Validation failed",
				"key": k,
				"value": attr_value,
				"schema": self._properties[k]
			})

def to_int(e):
	return e and int(e)
def to_str(e):
	return e and str(e)
def to_float(e):
	return e and float(e)

def item_validation(e, simple_types=(), complex_validations=(), nullable=True):
	if(e == None and not nullable):
		raise Exception("Cannot be none")

	valid = False
	if(simple_types and isinstance(e, simple_types)):
		valid = True
	if(complex_validations):
		for _complex_validation in complex_validations:
			if(_complex_validation(e)):
				valid = True
				break

	if(valid):
		return e

	if(nullable):
		return None

	raise Exception("Cannot be none")


def array_validation(arr, simple_types=(), complex_validations=(), default=_OBJ_END_, mix=False, nullable=True):
	#sequece
	if(arr == None and not nullable):
		if(default != _OBJ_END_):
			return default
		raise Exception("Cannot be none")
	if(not isinstance(arr, list)):
		arr = json.loads(arr)

	filter_nones = False
	_prev_type = _OBJ_END_
	for i in range(len(arr)):
		valid = False
		e = arr[i]
		if(simple_types and isinstance(e, simple_types)):
			valid = True
		if(complex_validations):
			for _complex_validation in complex_validations:
				if(_complex_validation(e)):
					valid = True
		#check types should not mixed
		if(not mix): # single type, so check type matches with previous
			_cur_type = type(e)
			if(_prev_type == _OBJ_END_):
				_prev_type = _cur_type
			elif(_prev_type != _cur_type):
				valid = False
					
		if(not valid):
			arr[i] = None
			filter_nones = True

	if(filter_nones and not nullable): # null not allowed
		n = len(arr)
		i = n - 1
		while(i >= 0):
			if(arr[i] == None):
				arr.pop(i)
			i -= 1

	return arr


'''
	given any instance/class, returns schema, _validation function
'''
def schema(x):
	if(isinstance(x, type) and issubclass(x, Object) and x != Object):
		ret = schema.defs.get(x.__name__)
		if(ret):
			return {"schema": {"$ref": "#/definitions/" + x.__name__}}, x.validate

		x._validations = _validations = {}
		x._properties = _properties = {}
		x._required = _required = set()
		for k, _type in get_type_hints(x).items():
			is_required = True
			if(_type.__module__ == 'typing'
				and getattr(_type, "__origin__", None) == typing.Union
				and getattr(_type, "__args__", None)
			):
				_type = _type.__args__[0]
				is_required = False
			if(isinstance(_type, _Optional)):
				_type = _type._types[0]
				is_required = False
			_schema_and_validation = schema(_type)
			if(_schema_and_validation):
				_properties[k], _validations[k] = _schema_and_validation
				is_required and _required.add(k)
				_default = x.__dict__.get(k, _OBJ_END_)
				if(_default != _OBJ_END_ and not isinstance(_type, type)):
					_type._default = _default



		#create schema
		x._schema_ = ret = schema.defs[x.__name__] = {
				"type": "object",
			}
		_title = getattr(x, "_title_", None)
		_description = getattr(x, "_description_", None)
		if(_title):
			ret["title"] = _title
		if(_description):
			ret["description"] = getattr(x, "_description_", None)
				
		if(_properties):
			ret["properties"] = _properties

		if(_required):
			ret["required"] = list(_required)
		
		return x._schema_, x.validate

	#special for tuples and list
	elif(isinstance(x, (list, tuple))): # x = [int, str]->oneof, (int, str)->anyof
		is_nullable = False
		if(None in x):
			x.pop(x.index(None))
			is_nullable = True

		simple_types = [] # str, int, etc
		complex_validations = [] # array, object, SomeObject
		_schemas = []
		for _type in x: #
			_s, _v = schema(_type)
			_schemas.append(_s)

			if(_type in (int, str, float)):
				simple_types.append(_type)
			else:
				complex_validations.append(_type)

		ret = None
		if(len(x) == 1):
			ret = _schemas[0]
		elif(isinstance(x, list)):
			ret = {"oneOf": _schemas}
		else:
			ret = {"anyOf": _schemas}

		if(is_nullable):
			ret["nullable"] = True

		return ret, partial(
						item_validation,
						simple_types=tuple(simple_types),
						complex_validations=complex_validations,
						nullable=is_nullable
					)

	elif(x == int or x == Int):
		return {"type": "integer"}, to_int

	elif(x == float):
		return {"type": "number", "format": "float"}, to_float

	elif(x == str or x == Str):
		return {"type": "string"}, to_str

	elif(isinstance(x, Str)):
		return x._schema_, x.validate

	elif(isinstance(x, Int)):
		return x._schema_, x.validate

	elif(x == Array or x == list): # genric
		return {"type": "array"}, (lambda e: e if isinstance(e, list) else None)

	elif(x == Object or x == dict):
		return {"type": "object"}, (lambda e: e if isinstance(e, dict) else None)

	elif(isinstance(x, Array)): # Array(_types)
		return x._schema_, x.validate

	elif(isinstance(x, Object)): # Object(id=Array)
		return x._schema_, x.validate

	else:
		return None, None


schema.defs = {}

def all_subclasses(cls):
	return set(cls.__subclasses__()).union(
		[s for c in cls.__subclasses__() for s in all_subclasses(c)]
	)


schema.init = lambda: [schema(x) for x in all_subclasses(Object)]



from gevent.pywsgi import WSGIServer


def hello(environ, start_response):
    path = environ['PATH_INFO'] 
    if(path == '/test' and environ['REQUEST_METHOD'] == 'GET'):
        status = '200 OK'
        text = "<html><body>Hello World</body></html>"
    else:
        status = '404 Not Found'
        text = "Not Found"

    #print("0", path, "0")

    body = text.encode('utf-8')
    response_headers = [
        ('Content-type', 'text/plain; charset=utf-8'),
        ('Content-Length', str(len(body)))]
    start_response(status, response_headers)
    return [body]


WSGIServer(('0.0.0.0', 8081), hello, log=None).serve_forever()



# -*- coding: utf-8 -*-

'''
DynamoDB KeyConditionExpression and FilterExpression
http://boto3.readthedocs.io/en/stable/reference/customizations/dynamodb.html#ref-dynamodb-conditions
'''


from decimal import Decimal
from boto3.dynamodb.conditions import Key, Attr

from .errors import ValidationException
from .helpers import smart_str

__all__ = ['Expression']


class Expression(object):

    def set(self, value,
            set_path=None,
            attr_label=None,
            if_not_exists=None,
            list_append=None):
        '''
        parameters:
            - value: value
            - set_path: attr path if not use attr name
            - attr_label: string attr label ex: label=':p'
            - if_not_exists: string path ex: Price
            - list_append: (tuple) path, index
            ex: (#pr.FiveStar, -1) to last
                (#pr.FiveStar, 0) to first
        examples:
            Test(realname='gs', score=100).update(Test.order_score.set(100))
            Test(realname='gs', score=100).update(
                Test.order_score.set(5, label=':p')
            Test(realname='gs', score=100).update(
                Test.order_score.set(100, if_not_exists=('order_score', 50)))
                
                
            Test(realname='gs', score=100).update(
                Test.ids.set(100, list_append=('ids')))
            or
            Test(realname='gs', score=100).update(
                Test.ids.list_append(100))

        return exp, {label: value}
        '''
        path = attr_label or self.name
        label = ":{name}".format(name=path)
        attr_name = "#{name}".format(name=path)
        # ExpressionAttributeValues
        value = self.typecast_for_storage(value)
        
        eav = {label: value}
        ean = {}
        if if_not_exists:
            no_path, operand = if_not_exists, value
            if isinstance(operand, float):
                operand = Decimal(str(operand))
            eav[label] = operand
            ean[attr_name] = path
            exp = '{name} = if_not_exists({path}, {label})'.format(
                name=attr_name, path=no_path, label=label)
        elif list_append:
            list_path, index = list_append
            if index == 0:
                exp = "{path} = list_append({label}, {path})".format(
                    path=list_path, label=label)
            elif index == -1:
                exp = "{path} = list_append({path}, {label})".format(
                    path=list_path, label=label)
                        
            elif(index > 0):
                exp = "{path}[{index}] = {label}".format(
                    path=attr_name, label=label)

            else:
                raise ValidationException('index error')
        else:
            path = set_path or self.name
            attr_name = "#{name}".format(name=attr_label or path)
            ean[attr_name] = path
            exp = '{path} = {label}'.format(path=attr_name, label=label)
        exp_attr = {
            'name': ean, # name substitutions
            'value': eav # value substitutions
        }
        #print(exp , exp_attr)
        return exp, exp_attr, 'SET'

    def update_dict(self, _dict):
        name_substitutions = {}
        value_substitutions = {}
        is_first = True
        exp = ""
        for key, val in _dict.items():
            key_str = str(key)
            exp = exp + " %s %s.#key_%s = :val_%s"%("" if is_first else "," , self.name, key_str, key_str)
            name_substitutions["#key_" + key_str] = key
            value_substitutions[":val_" + key_str] = val
            is_first = False
        #print(exp, {"name": name_substitutions, "value": value_substitutions})
        return exp, {"name": name_substitutions, "value": value_substitutions}, 'SET'

    def remove_from_dict(self, keys):
        name_substitutions = {}
        value_substitutions = {}
        is_first = True
        exp = ""
        for key in keys:
            key_str = str(key)
            exp = exp + " %s %s.#key_%s "%("" if is_first else "," , self.name, key_str)
            name_substitutions["#key_" + key_str] = key
            is_first = False
        #print(exp, {"name": name_substitutions, "value": value_substitutions})
        return exp, {"name": name_substitutions, "value": value_substitutions}, 'REMOVE'



    def list_append(self, value, path=None, index=-1,
                    attr_label=None):
        path = attr_label or path or self.name
        label = ":{name}".format(name=path)
        attr_name = "#{name}".format(name=path)
        if index == 0:
            exp = "{path} = list_append({label}, {path})".format(
                path=attr_name, label=label)
        elif index == -1:
            exp = "{path} = list_append({path}, {label})".format(
                path=attr_name, label=label)
                                    
        elif(index>0):
            exp = "{path}[{index}] = {label}".format(
                path=attr_name, label=label, index=index)
        else:
            raise ValidationException('index error')
        exp_attr = {
            'value': {label: value},
            'name': {attr_name: path}
        }
        return exp, exp_attr, 'SET'

    def remove(self, path=None, indexes=None):
        '''
        parameters:
            path: attr path
            index: (list) index ex: [2, 4]
        '''
        exp = ''
        path = path or self.name
        attr_name = "#{name}".format(name=path)
        ean = {attr_name: path}
        if self.field_type == 'list':
            for index in indexes:
                sub_exp = '{name}[{index}]'.format(name=attr_name,
                                                   index=index)
                if not exp:
                    exp = '{sub_exp}'.format(sub_exp=sub_exp)
                else:
                    exp = '{exp}, {sub_exp}'.format(exp=exp,
                                                    sub_exp=sub_exp)
            return exp, {'name': ean}, 'REMOVE'
        else:
            exp = '{path}'.format(path=path)
            return exp, {}, 'REMOVE'

    def add(self, value, path=None, attr_label=None):
        '''
        support num and set
        ADD Price :n    price += n
        ADD Color :c
        '''
        if self.field_type not in ('integer', 'float', 'set', 'dict'):
            raise ValidationException('Incorrect data type, only [integer, float, set, dict]')
        exp_attr = {}
        if not path:
            attr_name = "#{name}".format(name=attr_label or self.name)
            exp_attr['name'] = {attr_name: self.name}
        else:
            attr_name = attr_label or path
        label = ":{name}".format(name=self.name)
        exp = '{name} {label}'.format(name=attr_name, label=label)
        exp_attr['value'] = {label: value}
        return exp, exp_attr, 'ADD'
    
    def delete(self, values, attr_label=None):
        '''
        delete from set , values is a set
        '''
        if self.field_type not in ('set'):
            raise ValidationException('Incorrect data type, only [set]')
        exp_attr = {}
        
        attr_name = "#{name}".format(name=attr_label or self.name)
        exp_attr['name'] = {attr_name: self.name}
        
        label = ":{name}".format(name=self.name)
        exp = '{name} {label}'.format(name=attr_name, label=label)
        exp_attr['value'] = {label: values}
        return exp, exp_attr, 'DELETE'

    
    

    def typecast_for_storage(self, value):
        return smart_str(value)

    def _expression_func(self, op, *values, **kwargs):
        # for use by index ... bad
        values = list(map(self.typecast_for_storage, values))
        self.op = op
        self.express_args = values
        use_key = kwargs.get('use_key', False)
        if self.hash_key and op != 'eq':
            raise ValidationException('Query key condition not supported')
        elif self.hash_key or self.range_key or use_key:
            use_key = True
            func = getattr(Key(self.name), op, None)
        else:
            func = getattr(Attr(self.name), op, None)
        if not func:
            raise ValidationException('Query key condition not supported')
        return self, func(*values), use_key

    def _expression(self, op, value):
        if self.use_decimal_types:
            value = Decimal(str(value))
        label = ':%s' % self.name
        exp = '{name} {op} {value}'.format(name=self.name, op=op, value=label)
        return exp, label, value

    def eq(self, value):  # ==
        # Creates a condition where the attribute is equal to the value.
        # Attr & Key
        return self._expression_func('eq', value)

    def ne(self, value):  # !=
        # Creates a condition where the attribute is not equal to the value
        # Attr
        return self._expression_func('ne', value)

    def lt(self, value):  # <
        # Creates a condition where the attribute is less than the value.
        # Attr & Key
        return self._expression_func('lt', value)

    def lte(self, value):  # <=
        # Creates a condition where the attribute is less than or
        # equal to the value.
        # Attr & Key
        return self._expression_func('lte', value)

    def gt(self, value):  # >
        # Creates a condition where the attribute is greater than the value.
        # Attr & Key
        return self._expression_func('gt', value)

    def gte(self, value):  # >=
        # Creates a condition where the attribute is greater than or equal to
        # the value.
        # Attr & Key
        return self._expression_func('gte', value)

    def between(self, low_value, high_value):
        # Creates a condition where the attribute is greater than or equal to
        # the low value and less than or equal to the high value.
        # Attr & Key
        return self._expression_func('between', low_value, high_value)

    def begins_with(self, value):
        # Creates a condition where the attribute begins with the value
        # Attr & Key
        return self._expression_func('begins_with', value)

    def is_in(self, value):
        # Creates a condition where the attribute is in the value
        # Attr
        if self.hash_key or self.range_key:
            # ValidationException
            raise ValidationException('Query key condition not supported')
        return self.name, Attr(self.name).is_in(value), False

    def contains(self, value):
        # Creates a condition where the attribute contains the value.
        # Attr
        if self.hash_key or self.range_key:
            # ValidationException
            raise ValidationException('Query key condition not supported')
        return self.name, Attr(self.name).contains(value), False

    def exists(self):
        # Creates a condition where the attribute exists.
        # Attr
        if self.hash_key or self.range_key:
            # ValidationException
            raise ValidationException('Query key condition not supported')
        return self.name, Attr(self.name).exists(), False

    def not_exists(self):
        # Creates a condition where the attribute does not exists.
        # Attr
        if self.hash_key or self.range_key:
            # ValidationException
            raise ValidationException('Query key condition not supported')
        return self.name, Attr(self.name).not_exists(), False



'''
Created on 22-Aug-2017

@author: abhinav
'''

# argparse==1.2.1
# gevent==1.2.1
# greenlet==0.4.12
# protobuf==3.2.0
# boto3==1.4.4


import gevent
import signal
import ujson as json
import re
import urllib.parse
import time
import functools
import traceback
import inspect
from urllib.parse import urlencode

from gevent.threading import Lock
from gevent.server import StreamServer
from requests_toolbelt.multipart import decoder

from . import config as blaster_config
from .config import IS_DEV
from .common_funcs_and_datastructures import SanitizedDict, get_mime_type_from_filename, DummyObject,\
	set_socket_options
from .utils import events
from .logging import LOG_ERROR, LOG_SERVER
from .schema import Object, schema as schema_func
from .websocket.server import WebSocketServerHandler


_is_server_running = True
default_stream_headers = {
	'Content-Type': 'text/html; charset=utf-8',
	'Strict-Transport-Security': 'max-age=31536000; includeSubDomains',
	'X-Content-Type-Options': 'nosniff',
	'vary': 'Accept-Encoding',
	'X-Frame-Options': 'SAMEORIGIN',
	'X-UA-Compatible': 'IE=Edge',
	'X-XSS-Protection': '0'
}

I_AM_HANDLING_THE_SOCKET = object() # handler takes care of status line, headers and body


REUSE_SOCKET_FOR_HTTP = object()
_OBJ_END_ = object()

########some random constants
HTTP_MAX_REQUEST_BODY_SIZE = 1 * 1024 * 1024 # 1 mb
HTTP_MAX_HEADERS_DATA_SIZE = 16 * 1024 # 16kb


class BufferedSocket():
	
	is_eof = False
	sock = None
	store = None

	def __init__(self, sock):
		self.sock = sock
		self.store = bytearray()
		
	def close(self):
		self.sock.close()
		self.is_eof = True

	def send(self, *_data):
		total_sent = 0
		for data in _data:
			if(isinstance(data, str)):
				data = data.encode()
			n = 0
			data_len = len(data)
			data_mview = memoryview(data)
			while(n < data_len):
				sent = self.sock.send(data_mview[n:])
				if(sent < 0):
					return sent # return failed
				n += sent
			total_sent += n
		return total_sent
				
	def recv(self, n):
		if(self.is_eof):
			return None
		if(self.store):
			ret = self.store
			self.store = bytearray()
			return ret
		return self.sock.recv(n)

	def recvn(self, n):
		if(self.is_eof):
			return None
		while(len(self.store) < n):
			data = self.sock.recv(4096)
			if(not data):
				self.is_eof = True
				break
			self.store.extend(data)

		#return n bytes for now
		ret = self.store[:n]
		#set remaining to new store
		self.store = self.store[n:]
		return ret
		
	#fails if it couldn't find the delimiter until max_size
	def readuntil(self, delimiter, max_size, discard_delimiter):
		if(self.is_eof):
			return None
		#check in store
		if(isinstance(delimiter, str)):
			delimiter = delimiter.encode()
		
		delimiter_len = len(delimiter)
		#scan the store until end, if not found extend and continue until store > max_size
		to_scan_len = len(self.store)
		i = 0 # how much we scanned already
		
		_store = self.store # get a reference
		while(True):
			if(i > max_size):
				self.is_eof = True
				return None
			if(i >= delimiter_len):
				j = 0
				lookup_from = i - delimiter_len
				while(j < delimiter_len and _store[lookup_from + j] == delimiter[j]):
					j += 1

				if(j == delimiter_len):
					#found
					ret = None
					if(discard_delimiter):
						ret = _store[:i - delimiter_len]
					else:
						ret = _store[:i]
					self.store = _store[i:] # set store to unscanned/pending
					return ret
			if(i >= to_scan_len):
				#scanned all buffer
				data = self.sock.recv(4096)
				if(not data):
					self.is_eof = True
					return None

				_store.extend(data) # fetch more data
				to_scan_len = len(_store)
			i += 1

	def __getattr__(self, key):
		ret = getattr(self.sock, key, _OBJ_END_)
		if(ret == _OBJ_END_):
			raise AttributeError()
		return ret

def get_chunk_size_from_header(chunk_header):
	ret = 0
	i = 0
	l = len(chunk_header)
	A = ord('A')
	Z = ord('Z')
	_0 = ord('0')
	_9 = ord('9')
	while(i < l):
		c = ord(chunk_header[i])
		if(c >= A and c <= Z):
			ret = ret * 16 + (10 + c - A)
		elif(c >= _0 and c <= _9):
			ret = ret * 16 + (c - _0)
		else:
			return ret
		i += 1
	return ret


####parse query string
def parse_qs_modified(qs, keep_blank_values=True, strict_parsing=False):
	_dict = {}
	for name, value in urllib.parse.parse_qsl(qs, keep_blank_values, strict_parsing):
		if name.endswith("[]"):
			_values = _dict.get(name)
			if(_values == None):
				_dict[name] = _values = []
			elif(not isinstance(_values , list)): # make it a list of values
				_dict[name] = _values = [_values]
			_values.append(value)
		else:
			_dict[name] = value
	return _dict



class HeadersDict(dict):
	def __getitem__(self, k):
		return super().__getitem__(k.lower())

	def get(self, k, default=None):
		header_value = super().get(k.lower(), _OBJ_END_)
		if(header_value == _OBJ_END_):
			return default
		return header_value


class Query(Object):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)

class Headers(Object):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)

class Cookie(Object):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)

class Body(Object):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)

#Array(str), Array((int, str), default=None), Array(Object), Array(Pet)
#Object(id=int, name=str ), class Pet: id = int, name = str, friend = Pet
class Request:
	#class level
	_before_hooks = []
	_after_hooks = []

	after = []
	__argument_creator_hooks = {}

	#instance level
	sock = None
	_post = None
	_get = None
	_cookies = None
	_headers = None
	#cookies to send to client
	_cookies_to_set = None
	#url data, doesn't contain query string
	path = None

	@classmethod
	def before(cls, func=None):
		if(func and callable(func)):
			#decorating
			Request._before_hooks.append(func)
			return func
		return Request._before_hooks

	@classmethod
	def after(cls, func=None):
		if(func and callable(func)):
			#decorating
			Request._after_hooks.append(func)
			return func
		return Request._after_hooks


	def __init__(self, buffered_socket):
		self._get = SanitizedDict() # empty params by default
		self._headers = HeadersDict()
		self.sock = buffered_socket
		self.ctx = DummyObject()

	#searches in post and get
	def get(self, key, default=None, **kwargs):
		val = _OBJ_END_
		#try post params first
		if(self._post):
			val = self._post.get(key, default=_OBJ_END_, **kwargs)
		#try get param next
		if(val == _OBJ_END_ and self._get):
			val = self._get.get(key, default=_OBJ_END_, **kwargs)

		if(val == _OBJ_END_):
			return default
		return val

	def __setitem__(self, key, val):
		self._get[key] = val

	def __getitem__(self, key):
		ret = self.get(key, default=_OBJ_END_)
		if(ret == _OBJ_END_):
			raise LookupError()
		return ret

	def __getattr__(self, key):
		ret = self.get(key, default=_OBJ_END_)
		if(ret == _OBJ_END_):
			raise AttributeError()
		return ret

	def GET(self, key=None, **kwargs):
		if(key == None):
			return self._get
		return self._get.get(key, **kwargs)

	def POST(self, key=None, **kwargs):
		if(key == None):
			return self._post
		if(self._post):
			return self._post.get(key, **kwargs)
		return None

	def HEADERS(self, key=None, default=None):
		if(key == None):
			return self._headers
		return self._headers.get(key, default=default)

	def COOKIES(self, key, cookie_value=_OBJ_END_):
		ret = self._cookies.get(key)
		if(cookie_value != _OBJ_END_):
			if(self._cookies_to_set == None):
				self._cookies_to_set = {}
			if(isinstance(cookie_value, dict)):
				_cookie_val = str(cookie_value.pop("value", "1"))
				if(cookie_value):
					_cookie_val += ";"
					#other key=val atrributes of cookie
					for k, v in cookie_value.items():
						_cookie_val += (" " + k)
						if(v is not True): # if mentioned as True just the key is suffice like httpsOnly, secureOnly
							_cookie_val += "=%s;"%(str(v))
				cookie_value = _cookie_val
			if(isinstance(cookie_value, str)):
				self._cookies_to_set[key] = cookie_value
		return ret

	def parse_request_body(self, post_data_bytes, headers):
		if(not post_data_bytes):
			return None
		self._post = _post_params = SanitizedDict()
		content_type_header = headers.get("Content-Type", "")
		if(headers and content_type_header.startswith("multipart/form-data")):
			for part in decoder.MultipartDecoder(post_data_bytes, content_type_header).parts:
				content_disposition = part.headers.pop(b'Content-Disposition', None)
				if(not content_disposition):
					continue
				attrs = {}
				for i in content_disposition.split(b';'):
					key_val = i.split(b'=')
					if(len(key_val) == 2):
						key = key_val[0].strip()
						val = key_val[1].strip(b'"\'').decode()
						attrs[key] = val
				name = attrs.pop(b'name', _OBJ_END_)
				if(name != _OBJ_END_):
					#if no additional headers, and no other attributes,
					#it's just a plain input field
					if(not part.headers and not attrs):
						_post_params[name] = part.text
					else: # could be a file or other type of data, preserve all data
						_post_params[name] = {
							"attrs": attrs,
							"data": part.content,
							"headers": part.headers
						}
		else:
			post_data_str = post_data_bytes.decode('utf-8')
			if(post_data_str[0] == '{'
				or content_type_header == "application/json"
			):
				_post_params.update(json.loads(post_data_str))
			else: # urlencoded form
				_post_params.update(parse_qs_modified(post_data_str))

		return self

	#{type: hook...} function to call to create that argument

	@classmethod
	def set_arg_type_hook(cls, _type, hook=None):
		if(hook == None):
			#using as decorator  @Request.set_arg_type_hook(User)
			def wrapper(func):
				cls.__argument_creator_hooks[_type] = func
				return func
			return wrapper

		cls.__argument_creator_hooks[_type] = hook

	#create the value of the argument based on type, default
	def make_arg(self, name, _type, default, validator):
		if(not _type):
			return self.get(name, default)
		if(_type == Request):  # request_params: Request
			return self
		elif(_type == Query):  # query: Query
			return self._get
		elif(_type == Headers):# headers: Headers
			return self._headers
		elif(_type == Body):# headers: Headers
			return self._post

		elif(isinstance(_type, Query)): # Query(id=str, abc=str)
			if(not _type._validations):
				return self._get
			ret = {}
			for k, attr_validation in _type._validations.items():
				ret[k] = attr_validation(self._get.get(k, escape_quotes=False))
			return DummyObject(ret)

		elif(isinstance(_type, Headers)): # Headers('user-agent')
			if(not _type._validations):
				return self._get
			ret = {}
			for k, attr_validation in _type._validations.items():
				ret.k = attr_validation(self._headers.get(k))
			return DummyObject(ret)

		elif(isinstance(_type, Body)): # Headers('user-agent')
			if(not _type._post):
				return None
			ret = {}
			for k, attr_validation in _type._validations.items():
				ret.k = attr_validation(self._post.get(k, escape_quotes=False))
			return DummyObject(ret)

		#type should be class at this point
		has_arg_creator_hook = Request.__argument_creator_hooks.get(_type) # if has an arg injector, use it
		if(has_arg_creator_hook):
			return has_arg_creator_hook(self)

		elif(issubclass(_type, Query)): # :LoginRequestQuery
			ret = _type()
			for k, attr_validation in _type._validations.items():
				setattr(ret, k, self._get.get(k, escape_quotes=False))
			return ret

		elif(issubclass(_type, Body)):
			ret = _type()
			for k, attr_validation in _type._validations.items():
				setattr(ret, k, self._post.get(k, escape_quotes=False))
			return ret

		ret = self.get(name)
		return validator(ret) if validator else ret # get or post

	def to_dict(self):
		return {"get": self._get, "post": self._post}


#contains all running apps
_all_apps = set()
'''
	@app.route('/index.html')
'''
class App:
	route_handlers = None
	request_handlers = None
	name = None
	stream_server = None

	def __init__(self, **kwargs):
		self.info = {k: kwargs.get(k, "-") for k in {"title", "description", "version"}}
		self.route_handlers = []
		self.request_handlers = []

	def route(
		self,
		regex,
		methods=None,
		title='',
		description='',
		max_body_size=None,
		before=None,
		after=None
	):
		methods = methods or ("GET", "POST")
		if(isinstance(methods, str)):
			methods = (methods,)

		def wrapper(func):
			#arguments order
			#all untyped arguments are ignored should possiblly come from regex groups
			#typed args should be constructed, some may be constructed by hooks
			#named arguments not typed as passed on as with defaults

			self.route_handlers.append({
				"regex": regex,
				"func": func,
				"methods": methods,
				"title": title,
				"description": description,
				"max_body_size": max_body_size,
				"after": [after] if callable(after) else list(after or []),
				"before": [before] if callable(before) else list(before or [])
			})
			#return func as if nothing's decorated! (so you can call function as is)
			return func

		return wrapper


	def start(self, port=80, handlers=[], **ssl_args):
		#sort descending order of the path lengths
		self.route_handlers.sort(key=functools.cmp_to_key(lambda x, y: len(y["regex"]) - len(x["regex"]))) # reverse sort by length
		
		#all additional handlers go the bottom of the current list
		handlers_with_methods = []
		for handler in handlers:
			if(isinstance(handler, (list, tuple))):
				if(len(handler) < 2 or len(handler) > 3):
					continue
				handlers_with_methods.append({"regex": handler[0], "func": handler[1], "methods": ("GET",)})
			elif(isinstance(handler, dict)):
				_handler = {"methods": ("GET", "POST")} # defaults
				_handler.update(handler)
				handlers_with_methods.append(_handler)

		#add a openapi.json route handler
		self.openapi = {"components": {"schemas": dict(schema_func.defs)}, "paths": {}, "openapi": "3.0.1", "info": self.info}
		if(IS_DEV):
			self.route_handlers.append({
				"regex": "/openapi\.json",
				"methods": ("GET", ),
				"func": lambda: (["Content-type: application/json"], json.dumps(self.openapi))
			})

		#add additional handlers specified at start
		self.route_handlers.extend(handlers_with_methods)

		#create all schemas
		schema_func.init()

		regexes_map = {} # cache/update/overwrite

		#iterate smaller path to larger to fill the request_handlers map
		#build request_handlers map => {regex: {GET: handler,...}...}
		for handler in reversed(self.route_handlers):
			regex = handler["regex"]

			#check and identify path params if any from regex
			handler["path_params"] = _path_params = {}
			if(isinstance(regex, str)):
				#special case where path params can be {:varname} {:+varname}
				#replacing regex with regex haha!
				_path_params.update({x: False for x in re.findall(r'\{([a-zA-Z_][a-zA-Z0-9_]*)\}', regex)}) # optional
				_path_params.update({x: True for x in re.findall(r'\{\+([a-zA-Z_][a-zA-Z0-9_]*)\}', regex)}) # required
				_path_params.update({x: True for x in re.findall(r'\{\*([a-zA-Z_][a-zA-Z0-9_]*)\}', regex)}) # required

				regex = re.sub(r'\{([a-zA-Z_][a-zA-Z0-9_]*)\}', '(?P<\g<1>>[^/]*)', regex)
				regex = re.sub(r'\{\+([a-zA-Z_][a-zA-Z0-9_]*)\}', '(?P<\g<1>>[^/]+)', regex)
				regex = re.sub(r'\{\*([a-zA-Z_][a-zA-Z0-9_]*)\}', '(?P<\g<1>>.+)', regex)

			existing_regex_method_handlers = regexes_map.get(regex)
			if(not existing_regex_method_handlers):
				regexes_map[regex] = existing_regex_method_handlers = {}

				#check for full path matches, if the given regex doesn't match or end
				if(not regex.startswith("^")):
					regex = "^" + regex
				if(regex.endswith("/")):
					regex += "?"
				if(not regex.endswith("$")):
					regex = regex + "$"

				self.request_handlers.append((re.compile(regex), existing_regex_method_handlers))

			for method in handler["methods"]:
				existing_regex_method_handlers[method.upper()] = handler

			#infer arguments from the typing
			func = handler["func"]
			original_func = getattr(func, "_original", func)
			func_signature = inspect.signature(original_func)
			args_spec = func_signature.parameters

			args = [] # typed or untyped
			kwargs = []
			_arg_already_exists = set()
			for arg_name, _def in args_spec.items():
				
				if(arg_name in _arg_already_exists):
					continue
				_arg_already_exists.add(arg_name)

				_type = _def.annotation if _def.annotation != inspect._empty else None
				if(_def.default == inspect._empty):
					if(not _type):
						if(arg_name not in _path_params): # check if it's a path param
							continue
						else:
							_type = str # default type for path params
					#name, type, default, validator
					args.append((arg_name, _type, None, schema_func(_type)[1]))
				else:
					#name, type, default, validator
					kwargs.append((arg_name, _type, _def.default, schema_func(_type)[1]))

			handler["args"] = args
			handler["kwargs"] = kwargs
			if(func_signature.return_annotation != inspect._empty):
				handler["return"] = func_signature.return_annotation

			#fix before and after functions at start, cannot be modified later
			handler["before"] = tuple(handler.get("before", []) + Request._before_hooks)
			handler["after"] = tuple(handler.get("after", []) + Request._after_hooks)

			#openapi docs
			self.generate_openapi_doc(handler)
		
		#longer paths match first, hence reversed
		self.request_handlers.reverse()

		class CustomStreamServer(StreamServer):
			def do_close(self, *args):
				return

			#lets handle closing ourself
			def wrap_socket_and_handle(self, client_socket, address):
				# used in case of ssl sockets
				return self.handle(self.wrap_socket(client_socket, **self.ssl_args), address)

		LOG_SERVER("server_start", port=port)
		self.stream_server = CustomStreamServer(('', port), handle=self.handle_connection, **ssl_args)
		#keep a track

		_all_apps.add(self)

	def generate_openapi_doc(self, handler):
		regex_path = handler["regex"]
		regex_path = regex_path.replace("\\.", ".").replace("{+", "{").replace("{*", "{") # openapi doesn't support yet
		self.openapi["paths"][regex_path] = _api = {}
		func = handler["func"]
		args = handler["args"]
		kwargs = handler["kwargs"]
		return_type = handler.get("return", None)

		for method in handler["methods"]:
			_parameters = []
			_body_classes = []
			_api[method.lower()] = _method_def = {
				"summary": handler.get("title", ""),
				"description": handler.get("description", "")
			}
			#add header, query, cookie params
			for arg in (args + kwargs):
				_type = arg[1]
				if(_type and isinstance(_type, type)):
					if(issubclass(_type, Query)):
						for _key, _schema in _type._schema_["properties"].items():
							_param = {
								"in": "query",
								"name": _key,
								"schema": _schema
							}
							if(_key in _type._required):
								_param["required"] = True

							_parameters.append(_param)

					if(issubclass(_type, Body)):
						_body_classes.append(_type)

				if(isinstance(_type, (Headers, Query, Cookie))):
					_in = "header" if isinstance(_type, Headers) else ("query" if isinstance(_type, Query) else "cookie")
					for _k, _s in _type._properties.items():
						_param = {
							"in": _in,
							"name": _k,
							"schema": _s
						}
						if(_key in _type._required):
							_param["required"] = True
							
						_parameters.append(_param)


			#add path params
			for _name, _required in handler.get("path_params", {}).items():
				_param = {
					"in": "path",
					"name": _name,
					"schema": {"type": "string"}
				}
				#this is tricky,
				_param["required"] = True
					
				_parameters.append(_param)

			if(_parameters):
				_method_def["parameters"] = _parameters

			if(_body_classes):
				_schema_name = ""
				if(len(_body_classes) == 1):
					_schema_name = getattr(func, "_original", func).__name__
				elif(len(_body_classes) > 1):
					#merge multiple body classes
					_schema_name = "And".join(sorted([getattr(func, "_original", func).__name__ for x in _body_classes]))
					_properties = {}
					for _key, _schema in _type._schema_["properties"].items():
						_properties[_key] = _schema

					#add schemas
					self.openapi["components"]["schemas"][_schema_name] = {"type": "object", "properties": _properties}

				_method_def["requestBody"] = {
					"content": {
						"application/json": {"schema": {"$ref": "#/components/schemas/" + _schema_name}},
						"application/x-www-form-urlencoded": {"schema": {"$ref": "#/components/schemas/" + _schema_name}}
					},
					"required": True
				}

			_method_def["responses"] = _reponse = {"200": {"description": "OK"}}

			if(return_type):
				_reponse["content"] = {
					"application/*": {
						"schema": {
							"$ref": "#/components/schemas/" + getattr(func, "_original", func).__name__
						}
					}
				}

	def serve(self):
		self.stream_server.serve_forever()

	def stop(self):
		self.stream_server.stop()
		_all_apps.remove(self)

	@classmethod
	def response_body_to_parts(cls, ret):
		status = response_headers = body = None
		if(isinstance(ret, tuple)):
			body_len = len(ret)
			if(body_len > 0):
				body = ret[-1]
			if(body_len > 1):
				response_headers = ret[-2]
			if(body_len > 2):
				status = ret[-3]
		else:
			body = ret

		return status, response_headers, body

	def process_http_request(self, buffered_socket):
		resuse_socket_for_next_http_request = True
		request_line = buffered_socket.readuntil('\r\n', 4096, True) # ignore request lines > 4096 bytes
		if(not request_line):
			return
		post_data = None
		request_params = Request(buffered_socket)
		cur_millis = int(1000 * time.time())
		request_type = None
		request_path = None
		headers = None
		try:
			request_line = request_line.decode("utf-8")
			request_type, _request_line = request_line.split(" ", 1)
			_http_protocol_index = _request_line.rfind(" ")
			request_path \
				= request_params.path \
				= _request_line[: _http_protocol_index]
			#http_version = _request_line[_http_protocol_index + 1:]

			query_start_index = request_path.find("?")
			if(query_start_index != -1):
				query_string = request_path[query_start_index + 1:]
				request_params._get.update(parse_qs_modified(query_string))

				#strip query string from path
				request_path \
					= request_params.path \
					= request_path[:query_start_index]

			#find the handler
			handler = None
			for regex, method_handlers in self.request_handlers:
				request_path_match = regex.match(request_path)
				
				if(request_path_match != None):
					handler = method_handlers.get(request_type)
					if(not handler):
						#perform GET call by default
						LOG_SERVER(
							"handler_method_not_found",
							request_type=request_type,
							msg="performing GET by default"
						)
						if(request_type != "GET"):
							handler = method_handlers.get("GET")
					if(handler):
						#found a handler
						break

			if(not handler):
				raise Exception("Method not found")

			#parse the headers
			headers = request_params._headers
			max_headers_data_size = handler.get("max_headers_data_size") or HTTP_MAX_HEADERS_DATA_SIZE
			_headers_data_size = 0
			while(_headers_data_size < max_headers_data_size): # keep reading headers
				data = buffered_socket.readuntil('\r\n', max_headers_data_size, True)
				if(data == b''):
					break
				elif(data == None):
					return # won't resuse socket

				header_name_value_pair = data.split(b':', 1)
				if(len(header_name_value_pair) == 2):
					header_name , header_value = header_name_value_pair
					headers[header_name.lower().decode()] = header_value.lstrip(b' ').decode()
				
				_headers_data_size += len(data)


			#check if there is a content length or transfer encoding chunked
			content_length = int(headers.get("Content-Length", 0))
			_max_body_size = handler.get("max_body_size") or HTTP_MAX_REQUEST_BODY_SIZE
			if(content_length > 0):
				if(content_length < _max_body_size):
					post_data = buffered_socket.recvn(content_length)

			else: # handle chuncked encoding
				transfer_encoding = headers.get("transfer-encoding")
				if(transfer_encoding and "chunked" in transfer_encoding):
					post_data = bytearray()
					chunk_size = get_chunk_size_from_header(buffered_socket.readuntil('\r\n', 512, True))# hex
					while(chunk_size > 0 and len(post_data) + chunk_size < _max_body_size):
						data = buffered_socket.recvn(chunk_size)
						post_data.extend(data)
						buffered_socket.readuntil('\r\n', 8, False) # remove the trailing \r\n
						#next chunk size
						chunk_size = get_chunk_size_from_header(buffered_socket.readuntil('\r\n', 512, True))# hex
					#as bytes


			func = handler.get("func")
			#process cookies
			cookie_header = headers.get("Cookie", None)
			decoded_cookies = {}
			if(cookie_header):
				_temp = cookie_header.strip().split(";")
				for i in _temp:
					decoded_cookies.update(parse_qs_modified(i.strip()))
			request_params._cookies = decoded_cookies
			#process cookies end
			request_params.parse_request_body(post_data, headers)

			handler_args = []
			handler_kwargs = {}

			#update any path matches if they exist
			path_params = request_path_match.groupdict()
			for path_param, path_param_value in path_params.items():
				if(path_param_value):
					request_params._get[path_param] = path_param_value

			#set a reference to handler
			request_params.handler = handler

			for _args in handler["args"]:
				handler_args.append(request_params.make_arg(*_args))

			for name, *_args in handler["kwargs"]:
				handler_kwargs[name] = request_params.make_arg(name, *_args)

			before_handling_hooks = handler.get("before")
			after_handling_hooks = handler.get("after")

			response_from_handler = None
			for before_handling_hook in before_handling_hooks: # pre processing
				response_from_handler = before_handling_hook(request_params)
				if(response_from_handler):
					break
			
			if(not response_from_handler):
				#if before handlers return already something
				response_from_handler = func(*handler_args, **handler_kwargs)
				
				for after_handling_hook in after_handling_hooks: # post processing
					response_from_handler = after_handling_hook(request_params, response_from_handler)

				#check and respond to keep alive
				resuse_socket_for_next_http_request = headers.get('Connection', "").lower() == "keep-alive"
				
				##app specific headers
				#handle various return values from handler functions

			status, response_headers, body = App.response_body_to_parts(response_from_handler)


			if(status or body != I_AM_HANDLING_THE_SOCKET):
				status = str(status) if status else '200 OK'
				buffered_socket.send(b'HTTP/1.1 ', status, b'\r\n')

			if(response_headers or body != I_AM_HANDLING_THE_SOCKET):
				if(isinstance(response_headers, list)):
					#send only the specified headers
					for header in response_headers:
						buffered_socket.send(header, b'\r\n')

				else:
					if(isinstance(response_headers, dict)):
						#mix default stream headers
						#add default stream headers
						for key, val in default_stream_headers.items():
							if(key not in response_headers):
								buffered_socket.send(key, b': ', val, b'\r\n')

					elif(response_headers == None): # no headers => use default stream headers
						response_headers = default_stream_headers

					#send all basic headers
					for key, val in response_headers.items():
						buffered_socket.send(key, b': ', val, b'\r\n')

				#send any new cookies set
				if(request_params._cookies_to_set):
					for cookie_name, cookie_val in request_params._cookies_to_set.items():
						buffered_socket.send(b'Set-Cookie: ', cookie_name, b'=', cookie_val, b'\r\n')

				#send back keep alive
				if(resuse_socket_for_next_http_request):
					buffered_socket.send(b'Connection: keep-alive\r\n')

				if(body == I_AM_HANDLING_THE_SOCKET):
					#close the headers
					buffered_socket.send(b'\r\n')
	
			if(body == I_AM_HANDLING_THE_SOCKET):
				return I_AM_HANDLING_THE_SOCKET

			else:
				#just a variable to track api type responses
				response_data = None
				if(isinstance(body, (dict, list))):
					response_data = body
					body = json.dumps(body)

				#encode body
				if(isinstance(body, str)):
					body = body.encode()

				#close headers and send the body data
				if(body):
					#finalize all the headers
					buffered_socket.send(b'Content-Length: ', str(len(body)), b'\r\n\r\n')
					buffered_socket.send(body)
				else:
					buffered_socket.send(b'Content-Length: 0', b'\r\n\r\n')

			LOG_SERVER("http", response_status=status, request_type=request_type , path=request_path, wallclockms=int(1000 * time.time()) - cur_millis)

		except Exception as ex:
			stacktrace_string = traceback.format_exc()
			body = None
			if(blaster_config.server_error_page):
				body = blaster_config.server_error_page(
					request_type,
					request_path,
					request_params,
					headers,
					stacktrace_string=stacktrace_string
				)
			if(not body):
				body = b'Internal server error'
			buffered_socket.send(
					b'HTTP/1.1 ', b'502 Server error', b'\r\n',
					b'Connection: close', b'\r\n',
					b'Content-Length: ', str(len(body)), b'\r\n\r\n',
					body
			)
			LOG_ERROR(
				"http_error",
				exception_str=str(ex),
				stacktrace_string=stacktrace_string,
				request_type=request_type,
				request_path=request_path,
				request_params=request_params.to_dict()
			)
			if(IS_DEV):
				traceback.print_exc()
			#BREAK THIS CONNECTION
		
		if(resuse_socket_for_next_http_request):
			return REUSE_SOCKET_FOR_HTTP

	'''
		all the connection handling magic happens here
	'''
	def handle_connection(self, socket, address):
		buffered_socket = BufferedSocket(socket)
		close_socket = True
		while(True):
			ret = self.process_http_request(buffered_socket)
			if(ret == REUSE_SOCKET_FOR_HTTP):
				continue # try again
			elif(ret == I_AM_HANDLING_THE_SOCKET):
				close_socket = False
			break

		if(close_socket):
			buffered_socket.close()


@events.register_as_listener("sigint")
def stop_all_apps():
	LOG_SERVER("server_info", data="exiting all servers")
	global _is_server_running
	
	_is_server_running = False
	
	events.broadcast_event("blaster_before_shutdown")

	for app in list(_all_apps):
		app.stop() # should handle all connections gracefully

	events.broadcast_event("blaster_after_shutdown")


# create a global app for generic single server use
_main_app = App(title="Blaster", description="Built for speed and rapid prototyping..", version="0.0.368")

#generic global route handler
route = _main_app.route

#generic glonal route handler
def start_server(*args, **kwargs):
	_main_app.start(*args, **kwargs)
	_main_app.serve()


def redirect_http_to_https():
	def redirect(path, request_params=None, user=None):
		host = request_params.HEADERS('host')
		if(not host):
			return "404", {}, "Not understood"
		return "302 Redirect", ["Location: https://%s%s?%s"%(host, request_params.path, request_params.query_string)], "Need Https"

	http_app = App("port=80")
	http_app.start(port=80, handlers=[("(.*)", redirect)])
	gevent.spawn(http_app.serve)
	return http_app


def is_server_running():
	return _is_server_running


#### additional utils


# File handler for blaster server
def static_file_handler(_base_folder_path_, default_file_path="index.html", file_not_found_page_cb=None):
	cached_file_data = {}
	gevent_lock = Lock()

	def file_handler(path, request_params: Request):
		if(not path):
			path = default_file_path
		#from given base_path
		path = _base_folder_path_ + str(path)
		file_data = cached_file_data.get(path, None)
		resp_headers = None

		if(not file_data or time.time() * 1000 - file_data[2] > 1000 if IS_DEV else 2 * 60 * 1000): # 1 millis
			gevent_lock.acquire()
			file_hash_key = path + urlencode(request_params._get)[:400]
			file_data = cached_file_data.get(file_hash_key, None)
			if(not file_data or time.time() * 1000 - file_data[2] > 1000): # 1 millis
				#put a lock here
				try:
					data = open(path, "rb").read()
					mime_type_headers = get_mime_type_from_filename(path)
					if(mime_type_headers):
						resp_headers = {
											'Content-Type': mime_type_headers.mime_type,
											'Cache-Control': 'max-age=31536000'
									}
					
					file_data = (data, resp_headers, time.time() * 1000)
					cached_file_data[file_hash_key] = file_data
				except Exception as ex:
					if(file_not_found_page_cb):
						file_data = (
										file_not_found_page_cb(path, request_params=None, headers=None, post_data=None),
										None,
										time.time() * 1000
									)
					else:
						file_data = ("--NO-FILE--", None, time.time() * 1000)
					
			gevent_lock.release()
					
		data, resp_headers, last_updated = file_data
		return resp_headers, data
	
	return file_handler

####################Auto api response utilities
class AutoApiResponse:
	data = None
	templates = None
	frame_template = None
	meta_tags = None
	return_type_param = None

	def __init__(self, data, template=None,
			frame_template=None, templates=None, meta_tags=None,
			return_type_param="return_type"
		):
		self.data = data
		self.frame_template = frame_template
		self.meta_tags = meta_tags
		self.templates = {"content": template}
		self.return_type_param = return_type_param
		if(templates):
			self.templates.update(templates)

def auto_api_response(request_params, response):
	(status, headers, body) = App.response_body_to_parts(response)
	if(isinstance(body, AutoApiResponse)):
		#just keep a reference as we overwrite body variable
		blaster_response_obj = body
		#set body as default to repose_obj.data
		body = blaster_response_obj.data
		return_type = request_params.get(blaster_response_obj.return_type_param)
		#standard return_types -> json, content, None(=>content inside frame)
		if(return_type != "json"):
			#if template file is given and body should be a dict for the template
			templates = blaster_response_obj.templates
			#if no template given or returns just the blaster_response_obj.data
			template_exists \
				= (return_type and templates.get(return_type))\
				or blaster_response_obj.templates.get("content")

			if(template_exists and isinstance(body, dict)):
				tmpl_rendered = template_exists.render(
					request_params=request_params,
					**body
				)

				frame_template = blaster_response_obj.frame_template
				#if there is a return type, just retutn the template
				#render the full frame
				if(return_type or not frame_template):
					body = tmpl_rendered # partial

				else: # return type is nothing -> return full frame
					body = frame_template.render(
						body_content=tmpl_rendered,
						meta_tags=blaster_response_obj.meta_tags
					)
	return status, headers, body



#add a default arg hook
def _get_web_socket_handler(request_params):
	set_socket_options(request_params.sock)
	ws = WebSocketServerHandler(request_params.sock) # sends handshake automatically
	ws.do_handshake(request_params.HEADERS())
	return ws


Request.set_arg_type_hook(WebSocketServerHandler, _get_web_socket_handler)

#sigint event broadcast
gevent.signal_handler(signal.SIGINT, lambda : events.broadcast_event("sigint"))



'''
Created on 22-Aug-2017

@author: abhinav
'''
import blaster
from blaster.config import IS_DEV

## set the configs
from blaster.base import start_server, parse_qs_modified
import argparse
import ujson as json
import collections
from blaster.common_funcs_and_datastructures import LRUCache, create_signed_value,\
	get_random_id, decode_signed_value,\
	set_socket_options,\
	WebsocketConnection
from blaster.urllib_utils import get_data
import random
from blaster.websocket.server import WebSocketServerHandler
import time
from blaster.constants import TYPE_ERROR, TYPE_OK, LOG_TYPE_EXCEPTION
from blaster.base import route_handler, is_server_running, LOG_WARN
import socket
from blaster.common_funcs_and_datastructures import cur_ms
import gevent
from gevent.time import sleep
import urllib

blaster.config.secrets = {"v0": "asdsadkjhsakjdhjksadasdasd"}



'''
main server has all the sessions info
when other server starts , update_server_status  onto main server
create_session on main server will create on a minion server or on self
connect to the minion server for a particular session.
'''


#basic session descriptions
#session = {"id": , "addr": , "description": ,"image": , "type": , "last_few_chats": [] }
public_sessions = LRUCache(100000)
private_sessions = LRUCache(100000)


#TODO: get the public ip address somehow
self_server_addr = None

main_server_addr = None

#relevant on main server
servers = None


TYPE_SESSION = 10
TYPE_SESSIONS = 11
TYPE_USER = 12


random_user_names = ["Pandugadu", "Rgv", "Apple", "atalo ariti pandu", "erri pushpam", "naku koncham mental"]
def get_random_name(hash_val=None):
	if(hash_val):
		return random_user_names[hash(hash_val)%len(random_user_names)]
	return random.choice(random_user_names)




##app decorators

def need_user(func):
	def new_func(*args , **kwargs):
		if(kwargs.get("user", None)): # atleast one has to exist
			return func(*args, **kwargs)
		return {"type": TYPE_ERROR, "error_message": "no user ? darn !."}
		
	new_func._original = getattr(func, "_original", func)
	return new_func
	
def decode_user(func):
	def new_func(*args , **kwargs):
		request_params = kwargs.get("request_params", None)
		auth_key = request_params.get("auth_key", None)
		if(not auth_key):
			#try reading from headers
			cookie = request_params.HEADERS("Cookie", None)
			if(cookie):
				cookie = cookie.decode()
				_temp = cookie.strip('\r\n').split("; ")
				vals = {}
				for i in _temp:
					vals.update(parse_qs_modified(str(i)))
					
				auth_key = vals.get("auth_key", None)
		
		if(auth_key):
			user_data = decode_signed_value("user_data", auth_key, max_age_days=10)
			#_type = user_data.get("type",None)
			if(user_data):
				kwargs["user"] = json.loads(user_data)
			
			else: # simple user_id
				user_id = decode_signed_value("user_key", auth_key, max_age_days=10)
				if(user_id):
					kwargs["user"] = {"user_id": user_id, "name": get_random_name(user_id)}
			
				
			
			
		return func(*args, **kwargs)
	
	new_func._original = getattr(func, "_original", func)
	return new_func


def get_internal_auth_key():
	return create_signed_value("user_data", json.dumps({"user_id": "internal", "name": "don"}))


def update_stats_on_main():
	errors = 0
	sleep(60) # initial sleep
	while(is_server_running()):
		try:
			req = get_data("http://" + main_server_addr + "/update_server_stats?auth_key=" + get_internal_auth_key().decode() + "&server_addr=" + self_server_addr, post=json.dumps({"load": 0.0}))
			if(not req or json.loads(req.read())["type"] != TYPE_OK):
				errors += 1
			sleep(5 * 60) # 5 minutes sleep
		except Exception as ex:
			LOG_WARN(LOG_TYPE_EXCEPTION, ex="could not update on main server")
			sleep(1 * 60) # 1 minutes sleep


#master server
#show all recently active sessions #lru based
@route_handler('^/get_sessions')
@decode_user
@need_user
def get_sessions(request_params, user=None):
	recent_sessions = public_sessions.cache.items()[:-100]
	return {"type": TYPE_SESSIONS, "sessions": [session_obj for session_id, session_obj in recent_sessions]}


#get individual session data
@route_handler('^/get_session/(.*)')
@decode_user
@need_user
def get_session(session_id, request_params, user=None):
	session_id = request_params.get("session_id", session_id)
	session = None
	if(session_id and session_id.startswith("private__")):
		session = private_sessions.cache.get(session_id, None) # returns empty string
	else:
		session = public_sessions.cache.get(session_id, "") # returns empty string
	
	if(session):
		return {"type": TYPE_SESSION , "session" : session}
	else:
		return create_new_session._original(request_params, user, session_id=session_id)
		
	return {"type": TYPE_ERROR, "error_message": "not found"} # empty body

#probably called internally between servers to main server
@route_handler('^/update_session_data')
@decode_user
@need_user
def update_session_data(session_id, request_params, user=None):
	if(user["user_id"] != "internal"):
		return {"type": TYPE_ERROR, "error_message": "not authorized"} # empty body
	if(session_id):
		public_sessions.set("session_id", json.loads(request_params.POST())) # will update this blindly
		return {"type": TYPE_OK, "message": "session updated"} # empty body
	
	return {"type": TYPE_ERROR, "error_message": "not found"} # empty body

@route_handler('^/send_session_message/(.*)')
@decode_user
@need_user
def send_session_message(session_id, request_params, user=None):
	if(user["user_id"] != "internal"):
		return {"type": TYPE_ERROR, "error_message": "not authorized"} # empty body
	
	session_obj = session_data.exists(session_id)
	message_data = json.loads(request_params.POST())
	if(session_obj):
		session_connections = session_obj["connections"]
		session_connections_temp = session_obj["viewer_connections"]
		
		session_messages = session_obj["messages"]
		session_messages.append(message_data)
		# broadcast to all other connections in this session
		message_data_str = json.dumps(message_data)
		for conn, temp in session_connections.cache.items():
			try:
				conn.send(message_data_str)
			except Exception as ex:
				pass
		for conn, temp in session_connections_temp.cache.items():
			try:
				conn.send(message_data_str)
			except Exception as ex:
				pass

	
		return {"type": TYPE_OK, "message": "sent"} # empty body
	
	else:
		#main server
		if(session_id and session_id.startswith("private__")):
			session = private_sessions.cache.get(session_id, None) # returns empty string
		else:
			session = public_sessions.cache.get(session_id, "") # returns empty string
		
		#call on the server and return the response
		return get_data("http://" + session["addr"] + "/send_message/" + session_id + "?auth_key=" + get_internal_auth_key().decode(), post=request_params.POST()).read()
		



#update load etc
@route_handler('^/update_server_stats')
@decode_user
@need_user
def update_server_stats(request_params, user=None):
	server_addr = urllib.request.unquote(request_params.get("server_addr", None))
	if(user["user_id"] == "internal"):
		server_info = json.loads(request_params.POST().decode())
		server_info["last_updated"] = cur_ms()
		servers.set(server_addr, server_info)
		return {"type": TYPE_OK, "message": "updated"} # empty body
	
	return {"type": TYPE_ERROR, "error_message": "not authorized"} # empty body


'''on the main server'''  
@route_handler('^/create_new_session')
@process_post_params
@decode_user
@need_user
def create_new_session(request_params, user=None, session_id=None):
	#create on some random server a
	description = request_params.get("description", None)
	image = request_params.get("image", None)
	is_public = request_params.get("is_public", "true")
	available_servers = []
	for i, server in servers.cache.items():
		if(server["load"] < 1 and server["last_updated"] > cur_ms() - 10 * 60 * 1000): # updates less than 10 minutes
			available_servers.append(i)
	
	addr = random.choice(available_servers)
	
	if(not addr):
		return {"type": TYPE_ERROR, "error_message": "no address"} # empty body
	
	if(session_id):
		pass
	elif(is_public == "true"):
		session_id = get_random_id(10)
	else:
		session_id = "private__" + get_random_id(10)
		
	session_obj = {"session_id": session_id, "description" : description, "image": image, "type": 0, "addr": addr, "user": user}
	
	#update on that server
	req = get_data("http://" + addr + "/after_new_session_created?auth_key=" + get_internal_auth_key().decode(), post=json.dumps(session_obj))
	if(not req or json.loads(req.read())["type"] != TYPE_OK):
		return {"type": TYPE_ERROR, "error_message": "not found"} # empty body
		
	
	if(is_public == "true"):
		public_sessions.set(session_id, session_obj)
	else:
		private_sessions.set(session_id, session_obj)
		
	return {"type": TYPE_SESSION , "session" : session_obj}
			

#all sessions that exists on this server
session_data = LRUCache(10000)

@route_handler('^/after_new_session_created')
@decode_user
@need_user
def after_new_session_created(request_params, user=None):
	if(user["user_id"] != "internal"):
		return {"type": TYPE_ERROR, "error_message": "invalid user"} # empty body
	
	session_obj = json.loads(request_params.POST().decode())
	session_id = session_obj["session_id"] 
	existing_session_obj = session_data.exists(session_id)
	if(not existing_session_obj):
		session_data.set(session_id , session_obj)
		#create session fields
		session_obj["connections"] = LRUCache(500)
		session_obj["viewer_connections"] = LRUCache(500)
		session_obj["messages"] = collections.deque(maxlen=100)
	return {"type": TYPE_OK, "message": "updated"} # empty body
	

@route_handler('^/create_or_update_user')
@decode_user
def create_or_update_user(request_params, user=None):
	user_name = request_params.get("name", None) or get_random_name()
	user_id = None
	if(user):
		user_id = user["user_id"]
	return {"type": TYPE_USER, "user_name": user_name , "auth_key": create_signed_value("user_data", json.dumps({"name": user_name, "user_id" : user_id or get_random_id(10)}))}

#web socket based connection
@route_handler('^/join_session/(.*)')
@decode_user
@need_user
def join_session(session_id, request_params, user=None):
	session_id = session_id or request_params.get("session_id", None)
	session_obj = session_data.exists(session_id)
	if(not session_obj):
		return {"type": TYPE_ERROR, "error_message": "not found"} # empty body

	user_id = user["user_id"]
	session_connections = session_obj["connections"]
	session_connections_temp = session_obj["viewer_connections"]
	
	session_messages = session_obj["messages"]

	ws = WebSocketServerHandler(request_params.HEADERS()) # sends handshake automatically 
	set_socket_options(request_params.sock)
	
	def delete_connection(ex):
		session_connections.delete(ws)
	
	def on_connected():
		ws.conn_obj = conn = WebsocketConnection(ws, user_id)
		conn.is_viewer_list = True
		popped_connections = session_connections_temp.set(conn, True)
		for c , b in popped_connections:
			c.ws.close()
		
		#send last 100 messages
		initial_messages = json.dumps(session_messages)
		msg_to_send = {"type": 101, "payload": initial_messages}
		conn.send(json.dumps(msg_to_send), add_to_assumend_sent=False)
		
		
		
	def on_timeout():
		if(ws.conn_obj):
			ws.conn_obj.send(json.dumps({"type": 0, "timestamp": int(time.time() * 1000)}))
			return True
		return False
	
	def on_message():
		try:
			message_data = json.loads(ws.data)
			if(message_data.get("type", None) == 1):
				if(ws.conn_obj.is_viewer_list):
					session_connections_temp.delete(ws.conn_obj)
					ws.conn_obj.is_viewer_list = False
					popped_connections = session_connections.set(ws.conn_obj, True)
					for c , b in popped_connections:
							c.ws.close()
				
				
				message_data["src_id"] = user_id # user_id
				message_data["timestamp"] = int(time.time() * 1000) # user_id
				
				session_messages.append(message_data)
				#broadcast to all other connections in this session
				message_data_str = json.dumps(message_data)
				for conn, temp in session_connections.cache.items():
					if(conn != ws.conn_obj):
						try:
							conn.send(message_data_str)
						except Exception as ex:
							pass
				for conn, temp in session_connections_temp.cache.items():
					if(conn!=ws.conn_obj):
						try:
							conn.send(message_data_str)
						except Exception as ex:
							pass
		except Exception as ex:
			LOG_WARN(LOG_TYPE_EXCEPTION, exception=str(ex))
	
	ws.on_close = delete_connection
	ws.on_connected = on_connected
	ws.on_message = on_message
	ws.on_timeout = on_timeout
	ws.do_handshake(request_params.HEADERS())
	ws.start_handling()
	return





if __name__ == "__main__":
	
	parser = argparse.ArgumentParser(description='process arguments')
	parser.add_argument('--port',
		help='host port specifically')
	
	parser.add_argument('--addr',
		help='host name')
	
	parser.add_argument('--main_addr',
		help='main server host name')
	
	
	
	args = parser.parse_args()
	
	port = 8081
	if(args.port):
		port = int(args.port)
	
	#TODO: get the public ip address somehow
	self_server_addr = (args.addr or socket.gethostbyname(socket.gethostname())) + ":" + str(port)
	
	main_server_addr = args.main_addr or self_server_addr
	
	#relevant on main server
	servers = LRUCache(100, [(self_server_addr, {"load": 0.0, "last_updated": cur_ms()})]) # {server_id : address , load: }
	

	
	gevent.spawn(update_stats_on_main)
	
	start_server(port, handlers=[])



try:
    import simplejson as json
except ImportError:
    try:
        import ujson as json
    except ImportError:
        try:
            from django.utils import simplejson as json
        except ImportError:
            raise ImportError('A json library is required to use this python library')



from logging import DEBUG, INFO, WARN, ERROR
from .config import LOG_LEVEL
from datetime import datetime
import ujson as json

# CRITICAL-50 ERROR-40  WARNING-30  INFO-20  DEBUG-10  NOTSET-0

#more levels
SERVER_INFO = 31
APP_INFO = 32

log_level_to_names = {
	DEBUG: "DEBUG",
	INFO: "INFO",
	WARN: "WARNING",
	ERROR: "ERROR",
	SERVER_INFO: "SERVER",
	APP_INFO: "APP"
}

def LOG(level, log_type, **kwargs):
	if(level < LOG_LEVEL):
		return
	#TODO: send to UDP listener
	print(log_level_to_names.get(level, "SERVER"), log_type, datetime.now(), json.dumps(kwargs))

def LOG_SERVER(log_type, **kwargs):
	LOG(SERVER_INFO, log_type, **kwargs)

def LOG_APP_INFO(log_type, **kwargs):
	LOG(APP_INFO, log_type, **kwargs)

def LOG_DEBUG(log_type, **kwargs):
	LOG(DEBUG, log_type, **kwargs)

def LOG_INFO(log_type, **kwargs):
	LOG(INFO, log_type, **kwargs)

def LOG_WARN(log_type, **kwargs):
	LOG(WARN, log_type, **kwargs)

def LOG_ERROR(log_type, **kwargs):
	LOG(ERROR, log_type, **kwargs)



from blaster.common_funcs_and_datastructures import get_random_id, cur_ms
from blaster.config import IS_DEV
from blaster.mongo_orm import Connection, Model, Attribute, init_mongo_cluster


#step 1
class SalesAgent(Model):
	_collection_name_ = "domain_agents"

	company = Attribute(str)
	agent_user_id = Attribute(str)
	auth_level = Attribute(int, default=1)
	load = Attribute(int, default=0)
	data = Attribute(dict)
	created_at = Attribute(int , default=cur_ms)
	updated_at = Attribute(int, default=cur_ms)

	def before_update(self):
		self.updated_at = cur_ms()

	_index_ = [
				(company, agent_user_id),
				(company, load, {"unique": False})
			]



#call this function after you import all mongo db classes/models
def initialize_mongo():
	nodes = []
	if IS_DEV:
		nodes.append(Connection(host="localhost", port=27017, db_name="xyz"))
	else:
		nodes.append(Connection(host="mongodb://mongo-0.mongo.default.svc/", port=27017, db_name="xyz"))

	#check if connection exists
	nodes[0].db.client.server_info()

	init_mongo_cluster(nodes)


if __name__ == "__main__":
	initialize_mongo()

	for i in range(10):
		sales_agent = SalesAgent(company="example", agent_user_id=get_random_id(5), data={}, load=i).commit()


	sales_agent.data["says"] = "hello"
	sales_agent.commit()
	
	sales_agent.auth_level = 2
	sales_agent.commit()

	print(sales_agent.to_dict())

	#query single item only
	sales_agent = SalesAgent.get(company="example")
	print(sales_agent.to_dict())

	#direct mongo query
	sales_agents = SalesAgent.query({"company": "example", "agent_user_id": {"$in": ["abcd", "pqrs"]}})
	#wont have any
	for i in sales_agents:
		print(i.to_dict())

	#this rreturns a map
	sales_agents = SalesAgent.query({"company": "example"})
	for i in sales_agents:
		print(i.to_dict())



'''
Created on 05-Jun-2019

@author: abhinav
'''
import os
from ..connection_pool import get_from_pool, release_to_pool
from ..common_funcs_and_datastructures import MIME_TYPE_MAP


def generate_upload_url(file_name, base_folder, s3_connection_pool_name="s3" , bucket=None, redirect_url=None, mime_type=None):
    if(not mime_type):
        #just in case
        extension = os.path.splitext(file_name)[1][1:]
        mime_type = MIME_TYPE_MAP.get(extension)

    fields = {"acl": "public-read", "Content-Type": mime_type, "success_action_status": "200"}

    conditions = [
        {"acl": "public-read"},
        {"Content-Type": mime_type},
        {"success_action_status": "200"}
    ]
    if(redirect_url):
        conditions.append({"redirect": redirect_url})


    # Generate the POST attributes
    s3_key = base_folder + "/" + file_name

    s3 = get_from_pool(s3_connection_pool_name)
    post = s3.meta.client.generate_presigned_post(
        Bucket=bucket,
        Key=s3_key,
        Fields=fields,
        Conditions=conditions,
        ExpiresIn=604800
    )
    release_to_pool(s3, s3_connection_pool_name)

    return post





##verify multiple collection query works as expected
from blaster.common_funcs_and_datastructures import get_time_overlaps, retry
from datetime import datetime
from blaster.utils.data_utils import parse_string_to_units, parse_currency_string


def test_sanitized_dicts_list():
	from blaster.common_funcs_and_datastructures import SanitizedDict, SanitizedList
	sd = SanitizedDict(a="<a>", b="<b>")
	sd["c"] = "<c>"
	sd["d"] = {"e": "<e>", "f": "<f>"}

	for k, v in sd.items():
		print(k, v)
	for k, v in sd["d"].items():
		print(k, v)

	sl = SanitizedList(["<a>", "<b>"])
	sl.append({"c": "<c>", "d": "<d>"})
	sl.extend(["<e>", "<f>"])
	for i in sl:
		print(i)
	for k, v in sl[2].items():
		print(k, v)



print(
	get_time_overlaps(
		datetime(year=2021, month=10, day=1),
		datetime(year=2021, month=10, day=20),
		["Monday 10:30 - 12:30", "Tuesday 10:30 - 12:30"]
	)
)

print(
	get_time_overlaps(
		datetime(year=2021, month=10, day=1),
		datetime(year=2021, month=10, day=20),
		["5/12/2021 10:30 - 12:30"]
	)
)
print(
	get_time_overlaps(
		datetime(year=2021, month=10, day=1),
		datetime(year=2021, month=10, day=20),
		["5/10/2021 10:30 - 12:30"]
	)
)


@retry(10)
def can_retry():
	print("trying")
	raise Exception


#can_retry()



print(parse_string_to_units(".9 units"))

print(parse_string_to_units("0.9 units"))
print(parse_string_to_units("rs -1.9"))
print(parse_currency_string("INR 2000"))



'''
Created on 02-Jul-2019

@author: abhinav
'''
from blaster.urllib_utils import get_data
import json
import gevent
from blaster.websocket._core import create_connection
import time


auth_key1 = json.loads(get_data("http://localhost:8081/create_or_update_user").read())["auth_key"]
auth_key2 = json.loads(get_data("http://localhost:8081/create_or_update_user").read())["auth_key"]

print(auth_key1, auth_key2)
session_id = json.loads(get_data("http://localhost:8081/get_session/abhinav?auth_key="+auth_key1).read())["session"]["session_id"]
print(session_id)
   
ws1 = create_connection("ws://localhost:8081/join_session/"+session_id+"?auth_key="+auth_key1)

ws2 = create_connection("ws://localhost:8081/join_session/"+session_id+"?auth_key="+auth_key2)

ws3 = create_connection("ws://localhost:8081/join_session/"+session_id+"?auth_key="+auth_key2)

#initial messages
print(ws2.recv())
print(ws1.recv())
print(ws3.recv())


time.sleep(1)


ws1.send(json.dumps({'type':1, "payload":"Hello world" }))
time.sleep(0.1)
print(ws2.recv())
print(ws3.recv())

ws1.send_close()
ws2.send_close()


#! -*- coding: utf-8 -*-
import copy

from botocore.exceptions import ClientError
from blaster.connection_pool import use_connection_pool

from .table import Table
from .query import Query
from .fields import Attribute
from .errors import FieldValidationException, ValidationException, ClientException
from .helpers import get_items_for_storage
from .fields import ListField, SetField, DictField,\
    CharField


def _initialize_attributes(model_class, name, bases, attrs):
    """
    Initialize the attributes of the model.
    """
    model_class._attributes = {}

    # In case of inheritance, we also add the parent's
    # attributes in the list of our attributes
    for parent in bases:
        if not isinstance(parent, ModelMetaclass):
            continue
        for k, v in parent._attributes.items():
            model_class._attributes[k] = v

    for k, v in attrs.items():
        if isinstance(v, Attribute):
            model_class._attributes[k] = v
            v.name = v.name or k


def _initialize_indexes(model_class, name, bases, attrs):
    """
    Stores the list of indexed attributes.
    """
    model_class._local_indexed_fields = []
    model_class._local_indexes = {}
    model_class._global_indexes = {}
    model_class._hash_key = None
    model_class._range_key = None
    for parent in bases:
        if not isinstance(parent, ModelMetaclass):
            continue
        for k, v in parent._attributes.items():
            if v.indexed:
                model_class._local_indexed_fields.append(k)

    for k, v in attrs.items():
        if isinstance(v, Attribute):
            if v.indexed:
                model_class._local_indexed_fields.append(k)
                model_class._local_indexes[k] = '{table_name}_ix_{field}'.format(
                    table_name=name, field=k)
            elif v.range_key:
                model_class._range_key = k
            elif v.hash_key:
                model_class._hash_key = k
    if name not in ('ModelBase', 'Model') and not model_class._hash_key:
        raise ValidationException('hash_key is required')
    
    #global indexes
    #TODO: copy from all base classes
    if(model_class.__secondary_indexes__):
        for global_index_name, index_attrs in model_class.__secondary_indexes__.items():
            hash_key = index_attrs.get("hash_key")
            range_key = index_attrs.get("range_key",None)
            
            model_class._attributes[hash_key]._is_secondary_hash_key = global_index_name
            if(range_key):
                model_class._attributes[range_key]._is_secondary_range_key = global_index_name
                
            model_class._global_indexes[global_index_name] = index_attrs
    
    

class ModelMetaclass(type):

    """
    Metaclass of the Model.
    """

    __table_name__ = None
    __local_index__ = {}
    
    __secondary_indexes__ = None

    def __init__(cls, name, bases, attrs):
        super(ModelMetaclass, cls).__init__(name, bases, attrs)
        name = cls.__table_name__ or name
        _initialize_attributes(cls, name, bases, attrs)
        _initialize_indexes(cls, name, bases, attrs)


class ModelBase(object, metaclass=ModelMetaclass):

    _condition_expression = None
    _expression_attribute_values = None

    pending_updates = None
    
    def __init__(self):
        self.pending_updates = []

    @classmethod
    @use_connection_pool(db="dynamodb")
    def create(cls, db=None, overwrite=True, **kwargs):
        instance = cls(**kwargs)
        if not instance.is_valid():
            raise ValidationException(instance.errors)
        try:
            Table(instance, db).put_item(instance.item, overwrite=overwrite)
            item = cls(**instance.item)
            return item
        except ClientError as e:
            raise ClientException(e.response['Error']['Message'])

    def _prepare_update_item_params(self, **kwargs):
        params = kwargs
        if self._condition_expression:
            params['ConditionExpression'] = self._condition_expression
        if self._expression_attribute_values:
            params['ExpressionAttributeValues'] = self._expression_attribute_values
        return params

  
    @use_connection_pool(db="dynamodb")
    def update(self, *args, **kwargs):
        update_fields = {}
        db = kwargs.pop("db")
        ReturnValues = kwargs.pop('ReturnValues', 'ALL_NEW')
        ReturnConsumedCapacity = kwargs.pop('ReturnConsumedCapacity', 'NONE')
        params = self._prepare_update_item_params(
            ReturnValues=ReturnValues,
            ReturnConsumedCapacity=ReturnConsumedCapacity)
        if not self.validate_attrs(**kwargs):
            raise FieldValidationException(self._errors)
        for k, v in kwargs.items():
            field = self.attributes[k]
            update_fields[k] = field.typecast_for_storage(v)
        # use storage value
        item = Table(self, db).update_item(
            update_fields, *args, **params)
        value_for_read = self._get_values_for_read(item)
        for k, v in value_for_read.items():
            setattr(self, k, v)#set new values
        return self

    @classmethod
    @use_connection_pool(db="dynamodb")
    def get(cls, db=None, **primary_key):
        instance = cls(**primary_key)
        item = Table(instance, db).get_item()
        if not item:
            return None
        value_for_read = instance._get_values_for_read(item)
        return cls(**value_for_read)

    @classmethod
    @use_connection_pool(db="dynamodb")
    def batch_get(cls, *primary_keys, **kwargs):
        if(not primary_keys):
            return [] #no results
        instance = cls()
        projections = kwargs.get("projections", None)
        if(projections):
            kwargs["projections"] = [x.name for x in projections]
            
        items = Table(instance,kwargs.get("db")).batch_get_item(*primary_keys)
        results = []
        for item in items:
            value_for_read = instance._get_values_for_read(item)
            results.append(cls(**value_for_read))
        return results

    @classmethod
    @use_connection_pool(db="dynamodb")
    def batch_write(cls, kwargs, overwrite=False, db=None):
        '''
        kwargs: items list
        '''
        items_validated, instances = get_items_for_storage(cls, kwargs)
        instance = cls()
        Table(instance,db).batch_write(items_validated, overwrite=overwrite)
        return instances

    @classmethod
    @use_connection_pool(db="dynamodb")
    def batch_delete(cls, items=None, keys=None,db=None):
        instance = cls()
        return Table(instance,db).batch_delete(items, keys)

    #utility function
    def get_primary_key_dict(self):
        ret = {}
        ret[self.__class__._hash_key] = getattr(self, self.__class__._hash_key)
        if(self.__class__._range_key):
            ret[self.__class__._range_key] =  getattr(self, self.__class__._range_key)
        return ret
    
    def get_primary_key_tuple(self):
        ret = {}
        if(self.__class__._range_key):
            return (getattr(self, self.__class__._hash_key), getattr(self, self.__class__._range_key))
        return tuple(getattr(self, self.__class__._hash_key))
    
    @use_connection_pool(db="dynamodb")
    def delete(self , db=None):
        # delete an item
        return Table(self, db).delete_item()

    @classmethod
    @use_connection_pool(db="dynamodb")
    def delete_by_key(cls ,  db=None, **kwargs):
        # delete an item
        return Table(cls(), db).delete_item(key=kwargs)

    @classmethod
    def query(cls, *args):
        instance = cls()
        return Query(instance, *args)

    @classmethod
    @use_connection_pool(db="dynamodb")
    def scan(cls, db=None, **kwargs):
        instance = cls()
        return Table(instance, db).scan(**kwargs)

    @classmethod
    @use_connection_pool(db="dynamodb")
    def item_count(cls, db=None):
        instance = cls()
        return Table(instance, db).item_count()

    @use_connection_pool(db="dynamodb")
    def write(self,db=None):
        item = Table(self, db).put_item(self.item)
        return item

    def save(self, overwrite=False):
        if not self.is_valid():
            raise ValidationException(self.errors)
        self.write()
        return self

    


class Model(ModelBase):

    pending_updates = None
    
    def __init__(self, **kwargs):
        self.update_attributes(**kwargs)
        self.projections = []
        self.pending_updates = []

    def is_valid(self):
        """
        Returns True if all the fields are valid, otherwise
        errors are in the 'errors' attribute
        It first validates the fields (required, unique, etc.)
        and then calls the validate method.
        >>> from dynamodb import Model
        >>> def validate_me(field, value):
        ...     if value == "Invalid":
        ...         return (field, "Invalid value")
        ...
        >>> class Foo(Model):
        ...     bar = Attribute(validator=validate_me)
        ...
        >>> f = Foo()
        >>> f.bar = "Invalid"
        >>> f.save()
        False
        >>> f.errors
        ['bar', 'Invalid value']
        .. WARNING::
        You may want to use ``validate`` described below to validate your model
        """
        self.errors = []
        for field in self.fields:
            try:
                field.validate(self)
            except FieldValidationException as e:
                self.errors.extend(e.errors)
        self.validate()
        return not bool(self.errors)

    def validate_attrs(self, **kwargs):
        self._errors = []
        instance = copy.deepcopy(self)#copy all values to another object #discarded anyway
        for attr, value in kwargs.items():
            field = self.attributes.get(attr)
            if not field:
                raise ValidationException('Field not found: %s' % attr)
            setattr(instance, attr, value)
            try:
                field.validate(instance)
            except FieldValidationException as e:
                self._errors.extend(e.errors)
        return not bool(self._errors)

    def validate(self):
        """
        Overriden in the model class.
        The function is here to help you validate your model.
        The validation should add errors to self._errors.
        Example:
        >>> from zaih_core.redis_index import Model
        >>> class Foo(Model):
        ...     name = Attribute(required=True)
        ...     def validate(self):
        ...         if self.name == "Invalid":
        ...             self._errors.append(('name', 'cannot be Invalid'))
        ...
        >>> f = Foo(name="Invalid")
        >>> f.save()
        False
        >>> f.errors
        [('name', 'cannot be Invalid')]
        """
        pass


    
    #transaction.set_modify_conditions(Transaction.transaction_status.ne(TRANSACTION_STATUS_COMPLETED))
    def set_modify_conditions(self, *args):
        instance = self
        _condition_expression = None
        _expression_attribute_values = {}
        for field_inst, exp, is_key in args:
            if is_key:
                pass
            else:
                if not _condition_expression:
                    _condition_expression = exp
                else:
                    _condition_expression = _condition_expression & exp
        instance._condition_expression = _condition_expression
        instance._expression_attribute_values = _expression_attribute_values
        return instance


    def set_field(self, field_name, value, update_now=False):
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        if(isinstance(field, (CharField)) and not value):
            value = None
        if(value==None and isinstance(field, (ListField, SetField, DictField))): 
            #cannot set list or set field to NOne
            return self
        self.pending_updates.append(field.set(value))
        setattr(self, field.name, value)
        return self
    
    def add_to_field(self, field_name, value, update_now=False):
        if(not value): return self
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        self.pending_updates.append(field.add(value))
        cur_value = getattr(self, field.name)
        if(not cur_value):
            setattr(self, field.name, value)
        else:
            setattr(self, field.name, cur_value + value)
        return self
    
    def update_to_dict_field(self, field_name, _dict):
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
            
        cur_value = getattr(self, field.name)
        if(cur_value==None):
            cur_val = _dict
            setattr(self, field.name, cur_val)
            self.pending_updates.append(field.set(cur_val))
        else:
            cur_value.update(_dict)
            self.pending_updates.append(field.update_dict(_dict))
            
        return self
    
    def remove_from_dict_field(self, field_name, keys):
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
            
        cur_value = getattr(self, field.name)
        if(not cur_value):
            return self
        else:
            for i in keys:
                cur_value.pop(i, None)
            self.pending_updates.append(field.remove_from_dict(keys))
            
        return self
    
    def append_to_listfield(self, field_name, value, path=None, update_now=False):
        if(not value): return self
        if(not isinstance(value, list)): #should be a list
            value = [value]
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        cur_value = getattr(self, field.name)
        if(cur_value==None):#empty list field
            setattr(self, field.name, value)
            return self.set_field(field_name, value, update_now)#reset field
        
        self.pending_updates.append(field.list_append(value, path=path))
        cur_value.append(value)
        setattr(self, field.name, cur_value)
        return self
   
    def replace_by_index_in_listfield(self, field_name, index, value, path=None, update_now=False):
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        self.pending_updates.append(field.list_append(value, path=path, index=index))
        cur_value = getattr(self, field.name, [])
        if(cur_value):
            if(index>=len(cur_value)):
                cur_value.append(value)
            else:
                cur_value[index] = value
            setattr(self, field.name, cur_value)
        return self
    
    def remove_by_index_from_listfield(self, field_name, index, path=None, remove_from_list=False, update_now =False):
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        self.pending_updates.append(field.remove(path=path, indexes=[index]))
        cur_value = getattr(self, field.name)
        if(cur_value and remove_from_list):
            cur_value.pop(index)
        return self
            
    def add_to_setfield(self, field_name, values, update_now=False):
        if(not values or not isinstance(values, set)): return self
        
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        cur_value = getattr(self, field.name)
        if(not cur_value):
            setattr(self, field.name, values)
            return self.set_field(field_name, values, update_now)#reset field
            
        else:
            self.pending_updates.append(field.add(values))
            cur_value.update(values)
        return self

    def delete_from_setfield(self, field_name, values, path=None, update_now=False):
        if(not values or not isinstance(values, set)): return self
        
        if(isinstance(field_name, Attribute)):
            field = field_name
        else:
            field = self.attributes[field_name]
        self.pending_updates.append(field.delete(values))
        cur_value = getattr(self, field.name)
        if(cur_value):
            for i in values:
                cur_value.discard(i)            
            
        return self
    
    def commit_changes(self):
        #parse through all changes in fields
        ret = self
        if(self.pending_updates):
            ret = self.update(*self.pending_updates)#as list of updates
            del self.pending_updates[:]
            self._condition_expression = None
            self._expression_attribute_values = None
        return ret
        #TODO: iterate through all fields and compare and update accordingly, all at once



    def update_attributes(self, **kwargs):
        """
        Updates the attributes of the model.
        >>> class Foo(Model):
        ...    name = Attribute()
        ...    title = Attribute()
        ...
        >>> f = Foo(name="Einstein", title="Mr.")
        >>> f.update_attributes(name="Tesla")
        >>> f.name
        'Tesla'
        """
        params = {}
        attrs = self.attributes.values()
        for att in attrs:
            if att.name in kwargs:
                att.__set__(self, kwargs[att.name])
                params[att.name] = kwargs[att.name]
        return params

    @property
    def attributes(self):
        """Return the attributes of the model.
        Returns a dict with models attribute name as keys
        and attribute descriptors as values.
        """
        # print self._attributes
        return dict(self._attributes)

    @property
    def fields(self):
        """Returns the list of field names of the model."""
        return list(self.attributes.values())

    def _get_values_for_read(self, values):
        read_values = {}
        for att, value in values.items():
            if att not in self.attributes:
                continue
            descriptor = self.attributes[att]
            _value = descriptor.typecast_for_read(value)
            read_values[att] = _value
        return read_values

    def _get_values_for_storage(self):
        data = {}
        if not self.is_valid():
            raise FieldValidationException(self.errors)
        for attr, field in self.attributes.items():
            value = getattr(self, attr)
            if value is not None:
                data[attr] = field.typecast_for_storage(value)
        return data

    @property
    def item(self):
        return self._get_values_for_storage()
    
    def to_son(self):
        data = {}
        for attr, field in self.attributes.items():
            value = getattr(self, attr)
            if value is not None:
                data[attr] = field.typecast_for_read(value)
        return data
    
        
    



'''
Created on 04-Nov-2017

@author: abhinav
'''

import gevent
import os
import sys
import subprocess
import shlex
import collections
import random
import string
import socket
import struct
import fcntl
import html
from gevent import sleep
from functools import reduce as _reduce
from gevent.lock import BoundedSemaphore
from collections import namedtuple
from datetime import timezone
from datetime import datetime
from datetime import timedelta
import logging
import time
import hmac
import base64
import hashlib
from gevent.threading import Thread
from gevent.queue import Queue
import re
import six

import traceback

from .websocket._core import WebSocket
# from .config import IS_DEV
from .utils.xss_html import XssHtml
from .utils import events
from .logging import LOG_APP_INFO, LOG_WARN, LOG_DEBUG


_this_ = sys.modules[__name__]
_OBJ_END_ = object()

def cur_ms():
	return int(1000 * time.time())
#random id

#genetic LRU cache
class LRUCache:
	cache = None
	capacity = None

	def __init__(self, capacity, items=None):
		self.capacity = capacity
		self.cache = collections.OrderedDict(items or [])
		#in addition to recentness of use

	def exists(self, key):
		return self.cache.get(key, None)
		
	def get(self, key, default=None):
		try:
			#remove and reinsert into
			#ordered dict to move to recent
			value = self.cache.pop(key)
			self.cache[key] = value
			return value
		except KeyError:
			return default

	def set(self, key, value):
		removed_entries = []
		try:
			self.cache.pop(key)
		except KeyError:
			#new entry so cleanup space if it's beyond capacity
			while(len(self.cache) >= self.capacity):
				removed_entries.append(self.cache.popitem(last=False))
		self.cache[key] = value
		
		return removed_entries
		
	def delete(self, key):
		return self.cache.pop(key, None)
	
	def clear(self):
		return self.cache.clear()
	
	def to_son(self):
		ret = {}
		for key, val in self.cache:
			ret[key] = val.to_son() if hasattr(val, "to_son") else val
		return ret

class ExpiringCache:
	cache = None
	capacity = None
	ttl = None

	#ttl in millis, 5 minutes default
	def __init__(self, capacity, items=None, ttl=5 * 60 * 1000):
		self.capacity = capacity
		self.cache = collections.OrderedDict()
		self.ttl = ttl
		if(items):
			for k, v in items:
				self.set(k, v)

	def get(self, key, default=None):
		try:
			#remove and reinsert into
			#ordered dict to move to recent
			cur_timestamp = cur_ms()
			timestamp_and_value = self.cache.get(key)
			if(not timestamp_and_value):
				return default

			timestamp, value = timestamp_and_value
			if(cur_timestamp < timestamp + self.ttl):
				return value
			else:
				#remove if expired
				self.cache.pop(key, None)

			return default
		except KeyError:
			return default


	def set(self, key, value):
		removed_entries = []
		try:
			self.cache.pop(key)
		except KeyError:
			#new entry so cleanup space if it's beyond capacity
			cur_timestamp = cur_ms()
			while(len(self.cache) >= self.capacity):
				_key, (timestamp, _value) = self.cache.popitem(last=False)
				if(cur_timestamp < timestamp + self.ttl):
					removed_entries.append((_key, _value))

		self.cache[key] = (cur_ms(), value)
		
		return removed_entries

	def exists(self, key):
		return self.cache.get(key, _OBJ_END_) != _OBJ_END_
		
	def delete(self, key):
		return self.cache.pop(key, None)
	
	def clear(self):
		return self.cache.clear()
	
	def to_son(self):
		ret = {}
		cur_timestamp = cur_ms()
		for key, _val in self.cache.items():
			timestamp, val = _val
			if(cur_timestamp < timestamp + self.ttl):
				ret[key] = val.to_son() if hasattr(val, "to_son") else val
		return ret


#get SON of the fields from any generic python object
def to_son(obj):
	ret = obj.__dict__
	for k in ret.keys():
		if(ret[k] == None):
			del ret[k]
	return ret


def from_kwargs(cls, **kwargs):
	ret = cls()
	for key in kwargs:
		setattr(ret, key, kwargs[key])
	return ret

def get_by_key_list(d , *keyList):
	for key in keyList:
		if(not d):
			return None
		d = d.get(key, None)
	return d

def date2string(date):
	return date.isoformat()

def date2timestamp(dt):
	# datetime to timestamp
	if not isinstance(dt, datetime):
		return dt
	if(six.PY34):
		return dt.replace(tzinfo=timezone.utc).timestamp()
	else:
		timestamp = time.mktime(dt.timetuple()) + dt.microsecond / 1e6
		return timestamp

def timestamp2date(timestamp):
	if not isinstance(timestamp, (int, float)):
		return timestamp
	date = datetime.utcfromtimestamp(timestamp)
	return date

def zrfill(tup, n):
	return tup + tuple(0 for x in range(n - len(tup)))

def zlfill(tup, n):
	return tuple(0 for x in range(n - len(tup))) + tup


DATE_REGEX = re.compile("(\d{1,2})[\/\-\.](\d{1,2})[\/\-\.](\d{4})")
TIME_REGEX = re.compile("(?:(0?[1-9]|1[012])(?::([0-5]\d))?(?::([0-5]\d))?\x20?([AP]M))|([01]\d|2[0-3])(:[0-5]\d){1,2}")
DAY_REGEX = re.compile("(mon|tues|wed(nes)?|thur(s)?|fri|sat(ur)?|sun)(day)?", re.IGNORECASE)

DAYS_OF_WEEK = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']
def get_time_overlaps(a, b, c: str, limit=10, partial=True):
	if(isinstance(a, int)):
		a = timestamp2date(a)

	if(isinstance(b, int)):
		b = timestamp2date(b)

	time_ranges = []
	for x in c:
		date_matches = [m for m in DATE_REGEX.findall(x)]
		time_matches = [tuple(filter(lambda x: x, map(lambda x: x.strip(":"), m))) for m in TIME_REGEX.findall(x)]
		day_matches = list(
			filter(
				lambda x: DAYS_OF_WEEK.index(x) > 0,
				[x[m.start():m.end()].lower() for m in DAY_REGEX.finditer(x) or []]
			)
		)

		offset_check_x = timedelta(seconds=0)
		offset_check_y = timedelta(seconds=0)
		
		if(time_matches):
			hours, mins, secs, am_pm = zrfill(time_matches[0], 4)
			if(am_pm and am_pm.lower() == "pm"):
				hours = int(hours or 0) + 12
			offset_check_x = timedelta(hours=int(hours or 0), minutes=int(mins or 0), seconds=int(secs or 0))

			if(len(time_matches) > 1):
				hours, mins, secs, am_pm = zrfill(time_matches[1], 4)

				if(am_pm and am_pm.lower() == "pm"):
					hours = (hours or 0) + 12
				offset_check_y = timedelta(hours=int(hours or 0), minutes=int(mins or 0), seconds=int(secs or 0))
			else:
				offset_check_y = offset_check_y + timedelta(hours=1)


		if(date_matches):
			day, month, year, *_ = map(lambda x: int(x) if x else 0, date_matches[0])
			if(year < 2000):
				year = 2000 + year
			time_range_start = datetime(year=year, month=month, day=day) + offset_check_x
			time_range_end = datetime(year=year, month=month, day=day) + offset_check_y
			if(len(date_matches) > 1):
				day, month, year, *_ = map(lambda x: int(x) if x else 0, date_matches[1])
				if(year < 2000):
					year = 2000 + year
				time_range_end = datetime(year=year, month=month, day=day) + offset_check_y
			time_ranges.append((time_range_start, time_range_end))

		elif(day_matches):
			start_day = DAYS_OF_WEEK.index(day_matches[0].lower())
			if(start_day == -1):
				return None
			t = a
			while(t < b and len(time_ranges) < limit):
				if(t.weekday() == start_day):
					t_start_of_day = datetime(year=t.year, month=t.month, day=t.day)
					time_ranges.append((t_start_of_day + offset_check_x, t_start_of_day + offset_check_y))
				t += timedelta(days=1)

		else:
			t = a
			while(t < b and len(time_ranges) < limit):
				t_start_of_day = datetime(a.year, a.month, a.day)
				time_ranges.append((t_start_of_day + offset_check_x, t_start_of_day + offset_check_y))
				t += timedelta(days=1)

	ret = []
	for start, end in time_ranges:
		if(not partial):
			if(start >= a and end <= b):
				ret.append((start, end))
		else:
			if(start >= a and start <= b):
				ret.append((start, (end if end < b else b)))
			elif(start <= a and a <= end):
				ret.append((a, (end if end < b else b)))

	return ret


###############EPOCH
EPOCH = timestamp2date(0)

def find_index(a_list, value):
	try:
		return a_list.index(value)
	except ValueError:
		return -1

def find_nth(haystack, needle, n):
	''' Find position of nth occurance in a string'''
	if(not haystack):
		return -1
	start = haystack.find(needle)
	while start >= 0 and n > 1:
		start = haystack.find(needle, start + len(needle))
		n -= 1
	return start


SOME_TIME_WHEN_WE_STARTED_THIS = 1471504855
SOME_TIME_WHEN_WE_STARTED_THIS_MILLIS_WITH_10 = 16111506425808

### this function is inspired from instagram engineering post on generating 64bit keys with 12 bit shard_id inside it
# __thread_data = LRUCache(10000)
def generate_64bit_key(shard_id): # shard id should be a 12 bit number
	# may be later use dattime
	millis_elapsed = int((time.time() - SOME_TIME_WHEN_WE_STARTED_THIS) * 1000)
	
	_id = (((1 << 41) - 1) & millis_elapsed) << 23 # 41 bits , clear 22 places for random id and shard_id
	
	_id |= (((1 << 12) - 1) & shard_id) << 11 # 12 bit shard id, on top
	

	#increment per thread number
	# thread_id = threading.current_thread().ident
	# thread_data = __thread_data.get(thread_id, None)
	# if(not thread_data):
	#     thread_data = [0]
	#     __thread_data.set(thread_id , thread_data)
	# thread_data[0] = (thread_data[0] + 1)%(1 << 11)
	random_11_bits = random.randrange(0, 1 << 11)
	_id |= (((1 << 11) - 1) & random_11_bits) # clear 12 places for random

	#shard_id|timestmap|random_number
	
	return _id

def int_to_bytes(number: int) -> bytes:
	return number.to_bytes(length=(8 + (number + (number < 0)).bit_length()) // 8, byteorder='big', signed=True)


BASE_62_LIST = string.ascii_uppercase + string.digits + string.ascii_lowercase
BASE_62_DICT = dict((c, i) for i, c in enumerate(BASE_62_LIST))

def str_to_int(string, reverse_base=BASE_62_DICT):
	length = len(reverse_base)
	ret = 0
	for i, c in enumerate(string[::-1]):
		ret += (length ** i) * reverse_base[c]

	return ret

def int_to_str(integer, base=BASE_62_LIST):
	if integer == 0:
		return base[0]

	length = len(base)
	ret = ''
	while integer != 0:
		ret = base[integer % length] + ret
		integer //= length

	return ret

def get_int_id():
	return int(time.time() * 10000 - SOME_TIME_WHEN_WE_STARTED_THIS_MILLIS_WITH_10)

def get_str_id():
	return int_to_str(get_int_id())


def is_almost_equal(a, b, max_diff=0.01):
	if(a == None and b == None):
		return True
	if(a == None or b == None):
		return False
	diff = abs(a - b)
	if(diff < max_diff):
		return False
	return True


##### protobuf + mysql utility functions
def list_to_protobuf(values, message):
	'''parse list to protobuf message'''
	if not values:
		return
	if isinstance(values[0], dict): # value needs to be further parsed
		for v in values:
			cmd = message.add()
			dict_to_protobuf(v, cmd)
	else: # value can be set
		message.extend(values)


#just like move constructor ;) , move values much faster than copies
def dict_to_protobuf(values, obj, transformations=None, excludes=None, preserve_values=True):
	if(not preserve_values):
		if(transformations):
			for k, func in transformations.items():
				values[k] = func(values.get(k, None))
				
		if(excludes):
			for exclude, flag in excludes.items():
				if(hasattr(values, exclude)):
					del values[exclude]
			
	for k, v in values.items():
		if(preserve_values):
			if(transformations and k in transformations):
				v = transformations[k](v)
							
			if(excludes and k in excludes):
				continue

		if hasattr(obj, k):
			if isinstance(v, dict): # value needs to be further parsed
				dict_to_protobuf(v, getattr(obj, k))
			elif isinstance(v, list):
				list_to_protobuf(v, getattr(obj, k))
			else: # value can be set
				if v: # otherwise default
					setattr(obj, k, v)

def get_mysql_rows_as_dict(res):
	
	rows_as_dict = []
	for row in res.rows:
		as_dict = {}
		for field, val in zip(res.fields, row):
			as_dict[field[0]] = val
		
		
		rows_as_dict.append(as_dict)
	
	return rows_as_dict

########################

def remove_duplicates(lst):
	exists = {}
	ret = []
	for i in lst:
		if(i in exists):
			continue
		exists[i] = True
		ret.append(i)
	return ret


def get_shard_id(_id):
	return (int(_id) & (((1 << 12) - 1) << 11)) >> 11
	
	
## move it to seperate module ?
## copied from internet sha1 token encode - decode module

if hasattr(hmac, 'compare_digest'):  # python 3.3
	_time_independent_equals = hmac.compare_digest
else:
	def _time_independent_equals(a, b):
		if len(a) != len(b):
			return False
		result = 0
		if isinstance(a[0], int):  # python3 byte strings
			for x, y in zip(a, b):
				result |= x ^ y
		else:  # python2
			for x, y in zip(a, b):
				result |= ord(x) ^ ord(y)
		return result == 0


def utf8(value) -> bytes:
	if(isinstance(value, bytes)):
		return value
	return value.encode("utf-8")

def create_signed_value(name, value, secret):
	timestamp = utf8(str(int(time.time())))
	value = base64.b64encode(utf8(value))
	signature = _create_signature(secret, name, value, timestamp)
	signed_value = b"|".join([value, timestamp, signature])
	return signed_value

def decode_signed_value(name, value, secret, max_age_days=-1):
	if not value:
		return None
	parts = utf8(value).split(b"|")
	if(len(parts) < 3):
		return None
	#check signature matches or not
	signature = _create_signature(secret, name, parts[0], parts[1])
	if not _time_independent_equals(parts[2], signature):
		return None
	if(max_age_days > 0):
		timestamp = int(parts[1])
		if timestamp < time.time() - max_age_days * 86400:
			LOG_APP_INFO("cookie", msg="Expired cookie %s"%value)
			return None
		if timestamp > time.time() + 31 * 86400:
			# _cookie_signature does not hash a delimiter between the
			# parts of the cookie, so an attacker could transfer trailing
			# digits from the payload to the timestamp without altering the
			# signature.  For backwards compatibility, sanity-check timestamp
			# here instead of modifying _cookie_signature.
			LOG_APP_INFO("cookie", msg="Cookie timestamp in future; possible tampering %s"%value)
			return None
	if parts[1].startswith(b"0"):
		logging.warning("Tampered cookie %r", value)
		return None
	try:
		return base64.b64decode(parts[0])
	except Exception:
		return None


def _create_signature(secret, *parts):
	hash = hmac.new(utf8(secret), digestmod=hashlib.sha1)
	for part in parts:
		hash.update(utf8(part))
	return utf8(hash.hexdigest())


"""Dependencies are expressed as a dictionary whose keys are items
	and whose values are a set of dependent items. Output is a list of
	sets in topological order. The first set consists of items with no
	dependences, each subsequent set consists of items that depend upon
	items in the preceeding sets.
"""
def toposort(data):

	# Special case empty input.
	if len(data) == 0:
		return

	# Copy the input so as to leave it unmodified.
	data = data.copy()

	# Ignore self dependencies.
	for k, v in data.items():
		v.discard(k)
	# Find all items that don't depend on anything.
	extra_items_in_deps = _reduce(set.union, data.values()) - set(data.keys())
	# Add empty dependences where needed.
	data.update({item: set() for item in extra_items_in_deps})
	while True:
		ordered = set(item for item, dep in data.items() if len(dep) == 0)
		if not ordered:
			break
		yield ordered
		data = {item: (dep - ordered)
			for item, dep in data.items() if item not in ordered
		}
	if len(data) != 0:
		exception_string = 'Circular dependencies exist among these items: {{{}}}'.format(
				', '.join([
					'{!r}:{!r}'.format(key, value) for key, value in sorted(data.items())
				])
		)
		raise Exception(exception_string)




# Console or Cloud Console.
def get_random_id(length=10, include_timestamp=True):
	'''returns a 10 character random string containing numbers lowercase upper case'''
	'''http://stackoverflow.com/questions/2257441/random-string-generation-with-upper-case-letters-and-digits-in-python'''

	key_str = ''.join(random.SystemRandom().choice(BASE_62_LIST) for _ in range(length)) 
	if(include_timestamp):
		key_str = key_str + ("%d" % time.time())
	#key_str = hashlib.md5(key_str).hexdigest()
	return key_str

# jump consistent hash function , jump_hash("Asdasd", 100) assigns to a bucket
def jump_hash(key, num_buckets):
	b, j = -1, 0
	key = int(hashlib.md5(key).hexdigest(), 16)
	while j < num_buckets:
		b = int(j)
		key = ((key * int(2862933555777941757)) + 1) & 0xFFFFFFFFFFFFFFFF
		j = float(b + 1) * (float(1 << 31) / float((key >> 33) + 1))
	return int(b)


number_regex = re.compile(r"([0-9\.]+)")
non_alpha_regex = re.compile("[^0-9a-zA-Z \.]", re.DOTALL)
def sanitize_string(text):
	return non_alpha_regex.sub("", text)


non_alpha_regex_2 = re.compile("[^0-9a-zA-Z]", re.DOTALL)
def sanitize_to_id(text):
	return non_alpha_regex_2.sub("", text.lower())


EMAIL_REGEX = re.compile('^[_a-z0-9-]+(\.[_a-z0-9-]+)*(\+[_a-z0-9-]+)?\@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,6})$')
def is_valid_email(email):
	return EMAIL_REGEX.match(email)
	

SANITIZE_EMAIL_REGEX_NO_PLUS_ALLOWED = re.compile("\+[^@]*")
def sanitize_email_id(email_id, plus_allowed=True):
	if(not email_id):
		return None
	email_id = email_id.strip().lower()
	if(not EMAIL_REGEX.match(email_id)):
		return None

	if(not plus_allowed):
		email_id = SANITIZE_EMAIL_REGEX_NO_PLUS_ALLOWED.sub("", email_id)

	return email_id


# first minus second
def list_diff(first, second):
	if(not first or not second):
		return first
	second = set(second)
	return [item for item in first if item not in second]

# what are added vs what need to be deleted from first
#no order
def list_diff2(first, second):
	if(not first and second):
		return [], second
	if(second == None):
		return first, []
	second = set(second)
	first = set(first)
	# to add and delete
	return [item for item in first if item not in second],\
		[item for item in second if item not in first]

# [TODO:] returns least operations to modify first-second
# hamiltonian distance
def list_diff3(first, second):
	if(not first and second):
		return [], second
	if(second == None):
		return first, []
	second = set(second)
	first = set(first)
	# to add and delete
	return [item for item in first if item not in second],\
		[item for item in second if item not in first]



# a dummy object with given keys,values
class DummyObject:
	entries = None

	def __init__(self, entries=None, **kwargs):
		self.entries = entries or {}
		if(kwargs):
			self.entries = _entries = dict(entries)
			_entries.update(kwargs)

	def __setattr__(self, key, val):
		if(self.entries):
			self.entries[key] = val
		else:
			super().__setattr__(key, val)

	def __getattr__(self, key):
		return self.entries.get(key)

	def get(self, key, default=None):
		return self.entries.get(key, default)

	def __getitem__(self, key):
		return self.entries.get(key)

	def __setitem__(self, key, val):
		self.entries[key] = val

	def to_dict(self):
		return self.entries


# fully qualified name of the object
def object_name(o):
	# o.__module__ + "." + o.__class__.__qualname__ is an example in
	# this context of H.L. Mencken's "neat, plausible, and wrong."
	# Python makes no guarantees as to whether the __module__ special
	# attribute is defined, so we take a more circumspect approach.
	# Alas, the module name is explicitly excluded from __qualname__
	# in Python 3.

	module = o.__class__.__module__
	if module is None or module == str.__class__.__module__:
		return o.__class__.__name__  # Avoid reporting __builtin__
	else:
		return module + '.' + o.__class__.__name__


# normalize a int/str to n number of characters
def normalize(val, n=13):
	val = str(val)
	length = len(val)
	if(n >= length):
		return ("0" * (n - length)) + val

	raise Exception('cannot have greater length, ensure about this')

# normalize a int/str to n number of characters 
# and clips of any characters more than n characters
def normalize_no_warn(val, n=13):
	val = str(val)
	length = len(val)
	if(n >= length):
		return ("0" * (n - length)) + val
	return val



def ltrim(_str, str_to_remove):
	if(str_to_remove and _str.startswith(str_to_remove)):
		return _str[len(str_to_remove):]
	return _str

def rtrim(_str, str_to_remove):
	if(str_to_remove and _str.endswith(str_to_remove)):
		return _str[:len(_str) - len(str_to_remove)]
	return _str

def trim(_str, str_to_remove):
	return rtrim(ltrim(_str, str_to_remove), str_to_remove)


#remove zeros in the front
#returns 0 if it's all zeroes
def de_normalize(id1):
	i = 0
	length = len(id1)
	while(i < length):
		if(id1[i] != '0'):
			break
		i = i + 1
	if(i >= length):
		return '0'
	return id1[i:]

def set_non_blocking(fd):
	"""
	Set the file description of the given file descriptor to non-blocking.
	"""
	flags = fcntl.fcntl(fd, fcntl.F_GETFL)
	flags = flags | os.O_NONBLOCK
	fcntl.fcntl(fd, fcntl.F_SETFL, flags)


#generic thread pool to submit tasks which will be
#picked up wokers and processed
class Worker(Thread):
	
	is_running = True

	""" Thread executing tasks from a given tasks queue """
	def __init__(self, tasks):
		Thread.__init__(self)
		self.tasks = tasks
		self.daemon = True

	def run(self):
		while self.is_running:
			#get a task from queue
			func, args, kargs = self.tasks.get()
			try:
				func(*args, **kargs)
			except Exception:
				# An exception happened in this thread
				traceback.print_exc()
			finally:
				# Mark this task as done, whether an exception happened or not
				self.tasks.task_done()

class ThreadPool:
	""" Pool of threads consuming tasks from a single queue """
	#queue of tasks
	tasks = None
	worker_threads = None

	def __init__(self, num_threads):
		self.tasks = Queue(num_threads)
		self.worker_threads = []
		for _ in range(num_threads):
			worker_thread = Worker(self.tasks)
			self.worker_threads.append(worker_thread)
			#start the worker thread
			worker_thread.start()

	def add_task(self, func, *args, **kargs):
		""" Add a task to the queue """
		self.tasks.put((func, args, kargs))

	def join(self):
		""" Wait for completion of all the tasks in the queue """
		self.tasks.join()
		#stop processing
		for worker_thread in self.worker_threads:
			worker_thread.is_running = False
		#join all threads
		for worker_thread in self.worker_threads:
			worker_thread.join()

def make_xss_safe(_html):
	if(not _html):
		return _html
	parser = XssHtml()
	parser.feed(_html)
	parser.close()
	return parser.getHtml()

#yeild batches of items from iterator
def batched_iter(iterable, n=1):
	current_batch = []
	for item in iterable:
		current_batch.append(item)
		if len(current_batch) == n:
			yield current_batch
			current_batch = []
	if current_batch:
		yield current_batch


##### custom containers ##########
#SanitizedList and SanitizedDict are used for HTML safe operation
#the idea is to wrap them to sanitizeContainers while inserting
#rather than retrieving

class SanitizedSetterGetter(object):
	def __getitem__(self, k, escape_quotes=True, escape_html=True):
		val = super().__getitem__(k)
		if(escape_html and isinstance(val, str)):
			#make it html safe
			return html.escape(val, quote=escape_quotes)
		return val

	def __setitem__(self, key, val):
		if(isinstance(val, dict)):
			new_val = SanitizedDict()
			for k, v in val.items():
				new_val[k] = v # calls __setitem__ nested way
			super().__setitem__(key, new_val)

		elif(isinstance(val, list)):
			new_val = SanitizedList()
			for i in val:
				new_val.append(i)
			super().__setitem__(key, new_val)
		else:
			super().__setitem__(key, val)

class SanitizedList(SanitizedSetterGetter, list):

	def __iter__(self):
		#unoptimized but for this it's okay, always returns sanitized one
		def sanitized(val):
			if(isinstance(val, str)):
				return html.escape(val, quote=True)
			return val
		return map(sanitized, list.__iter__(self))

	def at(self, k, escape_quotes=True, escape_html=True):
		self.__getitem__(k,
			escape_quotes=escape_quotes,
			escape_html=escape_html
		)

	def extend(self, _list):
		for val in _list:
			#calls __setitem__ again
			self.append(val)
		#allow chaining
		return self

	def append(self, val):
		if(isinstance(val, dict)):
			new_val = SanitizedDict()
			for k, v in val.items():
				new_val[k] = v # calls __setitem__ nested way
			super().append(new_val)

		elif(isinstance(val, list)):
			new_val = SanitizedList()
			for i in val:
				new_val.append(i)
			super().append(new_val)
		else:
			super().append(val)
		#allow chaining
		return self

#intercepts all values setting and
class SanitizedDict(SanitizedSetterGetter, dict):
	#can pass escape_html=false if you want raw data
	def get(self, key, default=None, escape_html=True, escape_quotes=True):
		try:
			val = self.__getitem__(
				key,
				escape_html=escape_html,
				escape_quotes=escape_quotes
			)
			return val
		except KeyError:
			return default

	def items(self):
		#unoptimized but for this it's okay, always returns sanitized one
		def sanitized(key_val):
			key, val = key_val
			if(isinstance(val, str)):
				return (key, html.escape(val, quote=True))
			return key_val
		return map(sanitized, dict.items(self))

	def update(self, another):
		for k, v in another.items():
			#calls __setitem__ again
			self[k] = v
		#allow chaining
		return self


class LowerKeyDict(dict):
	def __getitem__(self, k):
		return super().__getitem__(k.lower())

	def get(self, k, default=None):
		return super().get(k.lower(), default)


# milli seconds
# TCP_USER_TIMEOUT kernel setting
_tcp_user_timeout = 30 * 1000

def set_socket_options(sock):
	l_onoff = 1
	l_linger = 10 # seconds,
	sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
				struct.pack('ii', l_onoff, l_linger))# don't wait too long to close the connection, after a timeout, abruptly close the connection
	
	timeval = struct.pack('ll', 10, 0) # it can wait 10 seconds, if there is congestion on the network card to send data
	sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO, timeval)

	TCP_USER_TIMEOUT = 18 # after 18 seconds if there is no ack then we assume broken and close it
	sock.setsockopt(socket.SOL_TCP, TCP_USER_TIMEOUT, _tcp_user_timeout) # close means a close understand ?

#wraps send method of websocket which keeps a buffer of messages for 20 seconds
#if the connection closes, you can use them to resend
class WebsocketConnection(WebSocket):
	queue = None# to keep track of how many greenlets are waiting on semaphore to send
	msg_assumed_sent = None# queue for older sent messages in case of reset we try to retransmit
	ws = None
	lock = BoundedSemaphore()
	is_stale = False
	last_msg_recv_timestamp = None
	last_msg_sent_timestamp = None
	
	
	user_id = None
	is_viewer_list = False
	
	def __init__(self, ws, user_id):
		self.ws = ws
		self.queue = collections.deque()# to keep track of how many greenlets are waiting on semaphore to send
		self.msg_assumed_sent = collections.deque()# queue for older sent messages in case of reset we try to retransmit
		self.user_id = user_id

		self.last_msg_recv_timestamp = self.last_msg_sent_timestamp = time.time() * 1000
		
		
	def send(self, msg, ref=None, add_to_assumend_sent=True): # msg is only string data , #ref is used , just in case an exception occurs , we pass that ref 
		if(self.is_stale):
			raise Exception("stale connection")
		
		self.queue.append((ref, msg))
		if(self.lock.locked()):
			return
		
		self.lock.acquire()
		data_ref = None
		data = None
		try:
			while(not self.is_stale and len(self.queue) > 0):
				data_ref, data = self.queue.popleft() # peek
				self.ws.send(data) # msg objects only
				current_timestamp = time.time() * 1000
				self.last_msg_sent_timestamp = current_timestamp
				if(add_to_assumend_sent):
					while(len(self.msg_assumed_sent) > 0 and self.msg_assumed_sent[0][0] < current_timestamp - _tcp_user_timeout):
						#keep inly 20 seconds of previous data
						self.msg_assumed_sent.popleft()
					
					self.msg_assumed_sent.append((current_timestamp , data_ref, data))
				
		except Exception as ex:
			err_msg = "Exception while sending message to %s might be closed "%(self.user_id)
			self.is_stale = True
			raise Exception(err_msg)
			
		finally:
			self.lock.release()
		return
	
	
#upload related mime_types
TypeDescriptor = namedtuple('TypeDescriptor', ['mime_type'])
MIME_TYPE_MAP = dict(
	gif=TypeDescriptor(mime_type='image/gif'),
	webp=TypeDescriptor(mime_type='image/webp'),
	mp4=TypeDescriptor(mime_type='video/mp4'),

	jpeg=TypeDescriptor(mime_type='image/jpeg'),
	jpg=TypeDescriptor(mime_type='image/jpeg'),
	png=TypeDescriptor(mime_type='image/png'),
	json=TypeDescriptor(mime_type='application/json'),
	zip=TypeDescriptor(mime_type='application/zip'),
	mp3=TypeDescriptor(mime_type='audio/mpeg'),
	ogg=TypeDescriptor(mime_type='audio/ogg'),
	opus=TypeDescriptor(mime_type='audio/ogg'),
	wav=TypeDescriptor(mime_type='audio/wav'),
	m4a=TypeDescriptor(mime_type='audio/mp4'),
	aac=TypeDescriptor(mime_type='audio/aac'),
	docx=TypeDescriptor(mime_type='application/vnd.openxmlformats-officedocument.wordprocessingml.document'),
	pdf=TypeDescriptor(mime_type='application/pdf'),

	# variables should not start numerals
	gp=TypeDescriptor(mime_type='audio/3gp'),
	amr=TypeDescriptor(mime_type='audio/amr'),
	wma=TypeDescriptor(mime_type='audio/wma'),
	gpp=TypeDescriptor(mime_type='audio/3gpp'),
	flac=TypeDescriptor(mime_type='audio/flac'),
	
	css=TypeDescriptor(mime_type='text/css'),
	js=TypeDescriptor(mime_type='text/javascript')
	
)
#mime_type -> file extension map
MIME_TYPE_TO_EXTENSION = {v.mime_type: k for k, v in MIME_TYPE_MAP.items()}


def get_mime_type_from_filename(file_name):
	ext = os.path.splitext(file_name)[1][1:]
	return MIME_TYPE_MAP.get(ext, None)



def parse_cmd_line_arguments():
	from sys import argv
	args_map = {}
	i = 0
	num_args = len(argv)
	while(i < num_args):
		arg = argv[i]
		if(arg.startswith("--")):
			if("=" in arg):
				key, val = arg.split("=", 1)
				args_map[key.strip("-")] = val
			else:
				args_map[arg.strip("-")] = True
		elif(arg.startswith("-")):
			key = arg.strip("-")
			val = argv[i + 1] if (i + 1 < num_args) else True
			i += 1
			args_map[key] = val

		i += 1
	return args_map


def run_shell(cmd, output_parser=None, shell=False, max_buf=5000):

	state = DummyObject()
	state.total_output = ""
	state.total_err = ""

	#keep parsing output
	def process_output(proc_out, proc_in):
		while(state.is_running):
			_out = proc_out.read(1)
			if(not _out):
				break
			_out = _out.decode('utf-8', 'ignore')
			#add to our input
			state.total_output += _out
			if(len(state.total_output) > 2 * max_buf):
				state.total_output = state.total_output[-max_buf:]

			if(output_parser):
				#parse the output and if it returns something
				#we write that to input file(generally stdin)
				_inp = output_parser(state.total_output, state.total_err)
				if(_inp):
					proc_in.write(_inp)
			else:
				print(_out, end="", flush=True)

	def process_error(proc_err, proc_in):
		while(state.is_running):
			_err = proc_err.read(1)
			if(not _err):
				break
			_err = _err.decode('utf-8', 'ignore')
			#add to our input
			state.total_err += _err
			if(len(state.total_err) > 2 * max_buf):
				state.total_err = state.total_err[-max_buf:]
			if(output_parser):
				#parse the output and if it returns something
				#we write that to input file(generally stdin)
				_inp = output_parser(state.total_output, state.total_err)
				if(_inp):
					proc_in.write(_inp)
			else:
				print(_err, end="", flush=True)

	if(isinstance(cmd, str) and not shell):
		cmd = shlex.split(cmd)

	dup_stdin = os.dup(sys.stdin.fileno()) if shell else subprocess.PIPE
	proc = subprocess.Popen(
		cmd,
		stdin=dup_stdin,
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE,
		shell=shell,
		env=os.environ.copy()
	)
	state.is_running = True

	#process output reader
	output_parser_thread = Thread(
		target=process_output,
		args=(
			proc.stdout,
			proc.stdin
		)
	)
	#process err reader
	err_parser_thread = Thread(
		target=process_error,
		args=(
			proc.stderr,
			proc.stdin
		)
	)
	output_parser_thread.start()
	err_parser_thread.start()

	os.close(dup_stdin) if dup_stdin != subprocess.PIPE else None

	#just keep printing error
	#wait for process to terminate
	ret_code = proc.wait()
	state.return_code = ret_code
	state.is_running = False
	
	output_parser_thread.join()
	err_parser_thread.join()
	return state


#args is array of strings or array or array of words
#you can use this to return a bunch of strings to index
#in elasticsearch with key "search_words"
def condense_for_search(*args):
	global_word_map = {}
	for arg in args:
		if(not arg):
			continue
		word_map = {}
		if(isinstance(arg, str)):
			arg = arg.split()
		for word in arg:
			key = word[:5]
			existing_words_of_key = word_map.get(key)
			if(not existing_words_of_key):
				word_map[key] = existing_words_of_key = []
			existing_words_of_key.append(word)
		for _key, words in word_map.items():
			_words = global_word_map.get(_key)
			# when existing matching words list
			# has more than current arg, ignore
			if(_words and len(_words) > len(words)):
				continue
			global_word_map[_key] = words

	ret = []
	for key, vals in global_word_map.items():
		ret.extend(vals)

	return ret

#returns None when there are exceptions instead of throwing
def ignore_exceptions(*exceptions):
	def decorator(func):
		def new_func(*args, **kwargs):
			try:
				func(*args, **kwargs)
			except Exception as ex:
				for exception in exceptions:
					if(isinstance(ex, exception)):
						return None
				raise ex

		new_func._original = getattr(func, "_original", func)
		return new_func
	return decorator

#r etries all exceptions or specific given expections only
#backoff = 1 => exponential sleep
#max_time milliseconds for exception to sleep, not counts the func runtime
def retry(num_retries=2, exceptions=None, max_time=5000):
	num_retries = max(2, num_retries)
	sleep_time_on_fail = max_time / num_retries
	exceptions = set(exceptions) if exceptions else None

	def decorator(func):
		def new_func(*args, **kwargs):
			retry_count = 0
			while(retry_count < num_retries):
				try:
					return func(*args, **kwargs)
				except Exception as ex:
					if(exceptions and ex not in exceptions):
						raise ex

					LOG_WARN("retrying", func=func.__name__, exception=str(ex))
					sleep(sleep_time_on_fail / 1000)
				retry_count += 1
			return None

		new_func._original = getattr(func, "_original", func)
		return new_func
	return decorator



def original_function(func):
	_original = getattr(func, "_original", _OBJ_END_)
	if(_original != _OBJ_END_):
		return _original

	while(True):
		func_wrapped = getattr(func, "__wrapped__", _OBJ_END_)
		if(func_wrapped == _OBJ_END_):
			break
		func = func_wrapped

	return func

def empty_func():
	pass


#when server shutsdown
_joinables = []
def start_background_thread(func, args=(), kwargs=None):
	if(not start_background_thread.can_run):
		LOG_WARN("background_threads", msg="Cannot run background threads. Correct copepaths")
		return

	LOG_APP_INFO("background_threads", msg="starting background thread", func=func.__name__)
	_thread_to_start = Thread(
		target=func,
		args=args,
		kwargs=kwargs
	)
	_joinables.append(_thread_to_start)
	_thread_to_start.start()


start_background_thread.can_run = True

#partioned queues
_partitioned_background_task_queues = tuple(Queue() for i in range(4))
def _process_partitioned_task_queue_items(_queue):
	while start_background_thread.can_run or not _queue.empty():
		func, args, kwargs = _queue.get()
		func(*args, **kwargs)

#singleton
def __start_task_processors():
	#start threads to process entries in partitioned queues
	for _queue in _partitioned_background_task_queues:
		start_background_thread(_process_partitioned_task_queue_items, args=(_queue,))
	__start_task_processors.started = True


#set initial flag
__start_task_processors.started = False

def submit_background_task(partition_key, func, *args, **kwargs):
	
	#start processors if not started already
	if(not __start_task_processors.started):
		__start_task_processors()

	if(partition_key == None):
		partition_key = cur_ms() # choose a random key
	_partitioned_background_task_queues[hash(str(partition_key)) % len(_partitioned_background_task_queues)]\
		.put((func, args, kwargs))


#decorator/wrapper to fire and forget
def background_task(func):
	def wrapper(*args, **kwargs):
		#spawn the thread
		if(not start_background_thread.can_run):
			LOG_WARN("background_threads", msg="Cannot run background threads. Correct copepaths")
			return
		submit_background_task(None, func, *args, **kwargs)
		return True

	wrapper._original = getattr(func, "_original", func)
	return wrapper

@events.register_as_listener(["sigint", "blaster_atexit", "blaster_after_shutdown"])
def cleanup_before_exit():

	if(not start_background_thread.can_run):
		return # double calling function

	start_background_thread.can_run = False

	#push an empty function to queues to flush them off
	for _partitioned_task_queue in _partitioned_background_task_queues:
		_partitioned_task_queue.put((empty_func, [], {}))

	#reap all joinables
	for _joinable in _joinables:
		_joinable.join()
	_joinables.clear()
	LOG_APP_INFO("background_threads", msg="cleanedup")


#calls a function after the function returns given by argument after
def call_after_func(func):

	if(isinstance(func, str)):
		# after_func => take from args named by func
		def decorator(func):
			def new_func(*args, **kwargs):
				after_func = kwargs.pop(func, None)
				ret = func(*args, **kwargs)
				after_func and after_func()
				return ret

			new_func._original = getattr(func, "_original", func)
			return new_func
		return decorator

	else:
		def new_func(*args, after=None, **kwargs):
			ret = func(*args, **kwargs)
			after and after()
			return ret
			
		new_func._original = getattr(func, "_original", func)
		return new_func



#mongo orm
#setup: 2 replica sets
#intialize table with primary and multiple secondary shards

#insert 10000 documents
##1. verify all 10000 are split to shards
##2. verify secondary shards are also updated for all 10000 documents

##update 100 documents
##1. verify  updates are propagated to the secondary shards and are available for query



from flask import Flask, render_template
import meinheld
import ujson as json

app = Flask(__name__)

@app.route('/hello')
def index():
    return json.dumps({"hello": "world"})


meinheld.listen(("0.0.0.0", 8000))
meinheld.run(app)



'''
Created on 22-Aug-2017

@author: abhinav
'''
import blaster
from blaster.base import start_server, route,\
    Request, static_file_handler
import os

@route("^/test$")
def do_test(request_params: Request):
    return "<html><body>Hello World</body></html>"

@route("^/hello")
def say_hello(request_params: Request):
    return ["content-type: application/json"], {"hello": "world"}

##directly return an object , server will automatically convert to json
@route("^/say/(.*)$")
def do_test2(something, request_params: Request):
    return {"message" : something} # something is a byte array !sorry

##directly return an object , server will automatically convert to json
@route("^/echo_params/{:something}$")
def do_test3(request_params: Request):
    return {"request_params" : request_params._get}

@route("^/echo_params/{:+something}$")
def do_test_with_something(request_params: Request):
    return {"request_params" : request_params._get}

##FOR handling you post arguments as key=value&key2=value2 data
@route("^/echo_post_params")
def do_test4(request_params: Request):
    return {"request_params" : request_params}

##FOR handling you post as raw body
@route("^/echo_post_data")
def do_test5(request_params: Request):
    return {"post_data" : request_params._post}

##FOR handling you post as raw json body
@route("^/echo_post_data")
def do_test6(request_params: Request):
    return {"post_data_object" : request_params}

#respond custom headers
@route("^/custom_header$")
def do_test7(request_params: Request):
    return {"X-header": "something_random"}, {"post_data_object" : request_params}

@route("^/custom_override_all_header$")
def override_all_headers(request_params: Request):
    return ["X-header: something_random"], {"post_data_object" : request_params}

#custom status
@route("^/custom_override_all_header$")
def do_test8(request_params: Request):
    return "302 Redirect", {"Location": "https://google.com"}, "Redirecting you"




if __name__ == "__main__":
    start_server(8081, handlers=[
            ('^/(.*)', static_file_handler(os.path.dirname(os.path.abspath(__file__)) + "/web/"))  # can server files too
    ])



import traceback
from botocore.exceptions import ClientError
from blaster.connection_pool import use_connection_pool
from .table import Table


@use_connection_pool(db="dynamodb")
def create_table(model_cls, db=None, force=False):
    
    table_obj = Table(model_cls(), db)
    if(force):
        try:
            table_obj.delete()
            table_obj.table.wait_until_not_exists()
        except Exception as ex:
            print("table doesnt exist ", table_obj.instance, ex)
        
    
    try:
        if(not force):
            try:
                table_obj.table.load()
                print("skipping %s table already exists"%(table_obj.table.name))
                return # continue if table already exists
            except ClientError as ce:
                pass

        table_obj.create()
        table_obj.table.wait_until_exists() # should wait until exisits
    except Exception as ex:
        print("could not create ", table_obj.instance, ex)
        traceback.print_exc()
